import io
import tarfile
import time
import uuid
from contextlib import contextmanager

import docker
import structlog
from docker.errors import APIError, DockerException, NotFound
from docker.models.containers import Container
from requests.exceptions import RequestException

from . import config, exceptions, utils

__all__ = ["create", "start", "destroy", "run", "working_directory"]

logger = structlog.get_logger()

_SANDBOX_NAME_PREFIX = "epicbox-"


class Sandbox:
    """
    Represent a sandbox Docker container.

    It can be used as a context manager to destroy the container upon
    completion of the block.
    """

    def __init__(self, id_, container: Container, realtime_limit=None):
        self.id_ = id_
        self.container = container
        self.realtime_limit = realtime_limit

    def __enter__(self):
        return self

    def __exit__(self, *args):
        destroy(self)

    def __repr__(self):
        return f"<Sandbox: {self.id_} сontainer={self.container.short_id}>"


def create(
    profile_name, command=None, files=None, limits=None, workdir=None
) -> Sandbox:
    """
    Create a new sandbox container without starting it.

    :param str profile_name: One of configured profile names.
    :param str command: A command with args to run in the sandbox container.
    :param list files: A list of `{'name': 'filename', 'content': b'data'}`
        dicts which define files to be written to the working directory
        of the sandbox.
    :param dict limits: Specify time and memory limits for the sandboxed
        process.  It overrides the default limits from `config.DEFAULT_LIMITS`.
    :param workdir: A working directory created using `working_directory`
                    context manager.
    :return Sandbox: A :class:`Sandbox` object.

    :raises DockerError: If an error occurred with the underlying
                         docker system.
    """
    if profile_name not in config.PROFILES:
        raise ValueError(f"Profile not found: {profile_name}")
    if workdir is not None and not isinstance(workdir, _WorkingDirectory):
        raise ValueError(
            "Invalid 'workdir', it should be created using "
            "'working_directory' context manager"
        )
    sandbox_id = str(uuid.uuid4())
    profile = config.PROFILES[profile_name]
    command = command or profile.command or "true"
    command_list = ["/bin/sh", "-c", command]
    limits = utils.merge_limits_defaults(limits)
    c = _create_sandbox_container(
        sandbox_id,
        profile.docker_image,
        command_list,
        limits,
        workdir=workdir,
        user=profile.user,
        read_only=profile.read_only,
        network_disabled=profile.network_disabled,
    )
    if workdir and not workdir.node:
        node_name = utils.inspect_container_node(c)
        if node_name:
            # Assign a Swarm node name to the working directory to run
            # subsequent containers on this same node.
            workdir.node = node_name
            logger.info("Assigned Swarm node to the working directory", workdir=workdir)
    if files:
        _write_files(c, files)
    sandbox = Sandbox(sandbox_id, c, realtime_limit=limits["realtime"])
    logger.info("Sandbox created and ready to start", sandbox=sandbox)
    return sandbox


def _create_sandbox_container(
    sandbox_id,
    image,
    command,
    limits,
    workdir=None,
    user=None,
    read_only=False,
    network_disabled=True,
):
    name = _SANDBOX_NAME_PREFIX + sandbox_id
    mem_limit = str(limits["memory"]) + "m"
    volumes = (
        {
            workdir.volume: {
                "bind": config.DOCKER_WORKDIR,
                "mode": "rw",
            }
        }
        if workdir
        else None
    )
    ulimits = utils.create_ulimits(limits)
    environment = None
    if workdir and workdir.node:
        # Add constraint to run a container on the Swarm node that
        # ran the first container with this working directory.
        environment = ["constraint:node==" + workdir.node]

    docker_client = utils.get_docker_client()
    log = logger.bind(sandbox_id=sandbox_id)
    log.info(
        "Creating a new sandbox container",
        name=name,
        image=image,
        command=command,
        limits=limits,
        workdir=workdir,
        user=user,
        read_only=read_only,
        network_disabled=network_disabled,
    )
    try:
        c = docker_client.containers.create(
            image,
            command=command,
            user=user,
            stdin_open=True,
            environment=environment,
            network_disabled=network_disabled,
            name=name,
            working_dir=config.DOCKER_WORKDIR,
            volumes=volumes,
            read_only=read_only,
            mem_limit=mem_limit,
            # Prevent from using any swap
            memswap_limit=mem_limit,
            ulimits=ulimits,
            # limit pid
            pids_limit=limits["processes"],
            # Disable the logging driver
            log_config={"type": "none"},
        )
    except (RequestException, DockerException) as e:
        if isinstance(e, APIError) and e.response.status_code == 409:
            # This may happen because of retries, it's a recoverable error
            log.info("The container with the given name is already created", name=name)
            c = {"Id": name}
        else:
            log.exception("Failed to create a sandbox container")
            raise exceptions.DockerError(str(e))
    log.info("Sandbox container created", container=c)
    return c


def start(sandbox, stdin=None):
    """
    Start a created sandbox container and wait for it to terminate.

    :param Sandbox sandbox: A sandbox to start.
    :param bytes or str stdin: The data to be sent to the standard input of the
        sandbox, or `None`, if no data should be sent.

    :return dict: A result structure containing the exit code of the sandbox,
        its stdout and stderr output, duration of execution, etc.

    :raises DockerError: If an error occurred with the underlying
                         docker system.
    """
    if stdin:
        if not isinstance(stdin, (bytes, str)):
            raise TypeError("'stdin' must be bytes or str")
        if isinstance(stdin, str):
            stdin = stdin.encode()
    log = logger.bind(sandbox=sandbox)
    log.info("Starting the sandbox", stdin_size=len(stdin or ""))
    result = {
        "exit_code": None,
        "stdout": b"",
        "stderr": b"",
        "duration": None,
        "timeout": False,
        "oom_killed": False,
    }
    try:
        stdout, stderr = utils.docker_communicate(
            sandbox.container, stdin=stdin, timeout=sandbox.realtime_limit
        )
    except TimeoutError:
        log.info("Sandbox realtime limit exceeded", limit=sandbox.realtime_limit)
        result["timeout"] = True
    except (RequestException, DockerException, OSError) as e:
        log.exception("Sandbox runtime error")
        raise exceptions.DockerError(str(e))
    else:
        log.info("Sandbox container exited")
        state = utils.inspect_exited_container_state(sandbox.container)
        result.update(stdout=stdout, stderr=stderr, **state)
        if (
            utils.is_killed_by_sigkill_or_sigxcpu(state["exit_code"])
            and not state["oom_killed"]
        ):
            # SIGKILL/SIGXCPU is sent but not by out of memory killer
            result["timeout"] = True
    log.info("Sandbox run result", result=utils.truncate_result(result))
    return result


def destroy(sandbox):
    """Destroy a sandbox container with improved error handling."""
    try:
        # First try to kill the container explicitly
        try:
            sandbox.container.kill(signal="SIGKILL")
        except (RequestException, DockerException):
            # Container might already be gone or not running - that's fine
            pass

        # Brief pause to let Docker process the kill
        time.sleep(0.5)

        # Then remove the container without volumes first
        try:
            sandbox.container.remove(v=False, force=True)
        except (RequestException, DockerException) as e:
            logger.exception("Failed to remove container", sandbox=sandbox, error=e)
    except Exception as e:
        logger.exception("Error during sandbox destruction", sandbox=sandbox, error=e)


def run(profile_name, command=None, files=None, stdin=None, limits=None, workdir=None):
    """
    Run a command in a new sandbox container and wait for it to finish
    running.  Destroy the sandbox when it has finished running.

    The arguments to this function is a combination of arguments passed
    to `create` and `start` functions.

    :return dict: The same as for `start`.

    :raises DockerError: If an error occurred with the underlying
                         docker system.
    """
    with create(
        profile_name, command=command, files=files, limits=limits, workdir=workdir
    ) as sandbox:
        return start(sandbox, stdin=stdin)


def run_exec(sandbox: Sandbox, command=None):
    """
    Run a command in an existing, running sandbox container and wait for it to finish
    running.  Destroy the sandbox when it has finished running.

    The arguments to this function is a combination of arguments passed
    to `create` and `start` functions.

    :return dict: The same as for `start`.

    :raises DockerError: If an error occurred with the underlying
                         docker system.
    """
    assert sandbox.container.status == "running"

    return sandbox.container.exec_run(command)


class _WorkingDirectory:
    """
    Represent a Docker volume used as a working directory.

    Not intended to be instantiated by yourself.
    """

    def __init__(self, volume, node=None):
        self.volume = volume
        self.node = node

    def __repr__(self):
        if self.node:
            return f"<WorkingDirectory: {self.volume} node={self.node}>"
        return f"<WorkingDirectory: {self.volume}>"


@contextmanager
def working_directory():
    docker_client = utils.get_docker_client()
    volume_name = "epicbox-" + str(uuid.uuid4())
    log = logger.bind(volume=volume_name)
    log.info("Creating new docker volume for working directory")
    try:
        volume = docker_client.volumes.create(volume_name)
    except (RequestException, DockerException) as e:
        log.exception("Failed to create a docker volume")
        raise exceptions.DockerError(str(e))
    log.info("New docker volume is created")
    try:
        yield _WorkingDirectory(volume=volume_name, node=None)
    finally:
        # Try volume cleanup with retries
        max_retries = 3
        for attempt in range(max_retries):
            try:
                log.info(f"Removing docker volume (attempt {attempt+1}/{max_retries})")
                volume.remove()
                log.info("Docker volume removed successfully")
                break
            except docker.errors.APIError as e:
                # Check if this is a "volume in use" error
                if "volume is in use" in str(e) and attempt < max_retries-1:
                    log.warning(f"Volume in use, waiting before retry: {e}")
                    time.sleep(2 ** attempt)  # Exponential backoff
                    # Try to find and force remove containers using this volume
                    try:
                        _cleanup_containers_using_volume(volume_name)
                    except Exception as cleanup_error:
                        log.exception("Error during container cleanup", error=cleanup_error)
                else:
                    log.exception("Failed to remove the docker volume")
            except Exception as e:
                log.exception("Unknown error removing volume", error=e)
                break


def _cleanup_containers_using_volume(volume_name):
    """Force remove any containers still using the specified volume."""
    docker_client = utils.get_docker_client()
    containers = docker_client.containers.list(all=True)
    for container in containers:
        # Check if container uses this volume
        # (This would require inspecting container mounts)
        try:
            container.remove(force=True)
        except Exception:
            pass


def _write_files(container, files):
    """Write files to the working directory in the given container."""
    # Retry on 'No such container' since it may happen when the function
    # is called immediately after the container is created.
    # Retry on 500 Server Error when untar cannot allocate memory.
    docker_client = utils.get_docker_client(retry_status_forcelist=(404, 500))
    log = logger.bind(files=utils.filter_filenames(files), container=container)
    log.info("Writing files to the working directory in container")
    mtime = int(time.time())
    files_written = []
    tarball_fileobj = io.BytesIO()
    with tarfile.open(fileobj=tarball_fileobj, mode="w") as tarball:
        for file in files:
            if not file.get("name") or not isinstance(file["name"], str):
                continue
            content = file.get("content", b"")
            file_info = tarfile.TarInfo(name=file["name"])
            file_info.size = len(content)
            file_info.mtime = mtime
            tarball.addfile(file_info, fileobj=io.BytesIO(content))
            files_written.append(file["name"])
    try:
        docker_client.api.put_archive(
            container.id, config.DOCKER_WORKDIR, tarball_fileobj.getvalue()
        )
    except (RequestException, DockerException) as e:
        log.exception(
            "Failed to extract an archive of files to the working "
            "directory in container"
        )
        raise exceptions.DockerError(str(e))
    log.info(
        "Successfully written files to the working directory",
        files_written=files_written,
    )



import json
import math
import os
from unittest.mock import MagicMock

import pytest
from lmwrapper.huggingface_wrapper import get_huggingface_lm
from lmwrapper.openai_wrapper import OpenAiModelNames, get_open_ai_lm
from lmwrapper.structs import ChatGptRoles, LmPrediction

from synthegrator.code_problems import CodeProblem, CodeSolution
from synthegrator.code_solver import LmCodeSolverAutoRegressive
from synthegrator.environments import ProjectWorkingDirectory
from synthegrator.few_shotting import FewShotConfig, FewShotLibrary
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec
from synthegrator.lm_few_shotting_tools import FewShotRendererWrapper
from synthegrator.memory_fs import ProjectDir
from synthegrator.prompt_renderer_questions import (
    CallbackResponseParser,
    QuestionPromptRenderer,
)
from synthegrator.transformation_spec import (
    SolveStep,
    StsCodeClassificationQuestion,
    StsCodeQuestion,
    StsSelectorFullFile,
    StsSpecStatement,
    TransformationSpec,
)

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


def basic_question_renderer():
    code = "if __name__ == '__main__':\n    print(2 + 3)\n"
    wd = ProjectWorkingDirectory(ProjectDir.construct_with_files([("main.py", code)]))
    spec = TransformationSpec(
        statements=(
            StsSpecStatement(
                "main.py",
                StsSelectorFullFile(),
                StsCodeQuestion(
                    "What will the code print? Please give only the value that appears on standard out with no extra text.",
                ),
            ),
        ),
    )
    return CodeProblem(
        working_directory=wd,
        transformation_spec=spec,
    )


def basic_classification_question_problem():
    code = "if __name__ == '__main__':\n    print(2 + 3)\n"
    wd = ProjectWorkingDirectory(ProjectDir.construct_with_files([("main.py", code)]))
    spec = TransformationSpec(
        statements=(
            StsSpecStatement(
                "main.py",
                StsSelectorFullFile(),
                StsCodeClassificationQuestion(
                    "Which programming language is used?",
                    ("python", "java", "c++"),
                ),
            ),
        ),
    )
    problem = CodeProblem(
        working_directory=wd,
        transformation_spec=spec,
    )
    problem.known_solutions = [
        CodeSolution(
            solve_steps=[
                SolveStep(
                    "main.py",
                    spec.statements[0].static_mark_id,
                    "python",
                ),
            ],
            problem=problem,
            lang_spec=PythonLangSpec(),
        ),
    ]
    return problem


def basic_classification_question_problem_2():
    code = "if __name__ == '__main__':\n    print(2 + 3)\n"
    wd = ProjectWorkingDirectory(ProjectDir.construct_with_files([("main.py", code)]))
    spec = TransformationSpec(
        statements=(
            StsSpecStatement(
                "main.py",
                StsSelectorFullFile(),
                StsCodeClassificationQuestion(
                    "Which value will be printed?",
                    ("1", "5", "hello world", "2"),
                ),
            ),
        ),
    )
    problem = CodeProblem(
        working_directory=wd,
        transformation_spec=spec,
    )
    problem.known_solutions = [
        CodeSolution(
            solve_steps=[
                SolveStep(
                    "main.py",
                    spec.statements[0].static_mark_id,
                    "5",
                ),
            ],
            problem=problem,
            lang_spec=PythonLangSpec(),
        ),
    ]
    return problem


def test_simple():
    problem = basic_question_renderer()
    renderer = QuestionPromptRenderer()
    render = renderer.render(
        problem,
        MagicMock(),
    )
    assert (
        render.prompt.text == "```python\n"
        "if __name__ == '__main__':\n"
        "    print(2 + 3)\n"
        "```\n"
        "Question: What will the code print? Please give only the value that appears on standard out with no extra text.\n"
        "Answer:"
    )


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="No api in github actions")
def test_simple_solve():
    problem = basic_question_renderer()
    renderer = QuestionPromptRenderer()
    parser = CallbackResponseParser()
    lm = get_open_ai_lm()
    solver = LmCodeSolverAutoRegressive(
        lm,
        prompt_renderer=renderer,
        response_parser=parser,
    )
    solution = solver.solve(problem)
    assert len(solution.solve_steps) == 1
    assert solution.solve_steps[0].path == "main.py"
    assert solution.solve_steps[0].value.strip() == "5"


def test_simple_solve_mocked():
    problem = basic_question_renderer()
    renderer = QuestionPromptRenderer()
    parser = CallbackResponseParser()
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo_instruct)
    lm.predict = MagicMock(return_value=LmPrediction("5\n", None, None))
    solver = LmCodeSolverAutoRegressive(
        lm,
        prompt_renderer=renderer,
        response_parser=parser,
    )
    solution = solver.solve(problem)
    assert len(solution.solve_steps) == 1
    assert solution.solve_steps[0].path == "main.py"
    assert solution.solve_steps[0].value.strip() == "5"


def test_classication_render():
    problem = basic_classification_question_problem()
    renderer = QuestionPromptRenderer()
    render = renderer.render(
        problem,
        MagicMock(),
    )
    # fmt: off
    assert render.prompt.text == (
        "```python\n"
        "if __name__ == '__main__':\n"
        "    print(2 + 3)\n"
        "```\n"
        "Question: Which programming language is used? Answer with only the letter corresponding to the best answer.\n"
        "A) python\n"
        "B) java\n"
        "C) c++\n"
        "Answer:"
    )
    # fmt: on


def test_classication_render_few_shot():
    problem = basic_classification_question_problem()
    few_shot_config = FewShotConfig(
        FewShotLibrary([basic_classification_question_problem_2()]),
        num_examples=1,
        change_examples_between_solves=False,
    )
    renderer = QuestionPromptRenderer()
    parser = CallbackResponseParser()
    renderer = FewShotRendererWrapper(
        renderer,
        parser,
        few_shot_config,
    )
    lm = MagicMock()
    lm.is_chat_model = False
    render = renderer.render(
        problem,
        lm,
    )
    print(render.prompt.text)
    # fmt: off
    assert render.prompt.text == (
        "```python\n"
        "if __name__ == '__main__':\n"
        "    print(2 + 3)\n"
        "```\n"
        "Question: Which value will be printed? Answer with only the letter corresponding to the best answer.\n"
        "A) 1\n"
        "B) 5\n"
        "C) hello world\n"
        "D) 2\n"
        "Answer: B"
        "\n###\n"
        "```python\n"
        "if __name__ == '__main__':\n"
        "    print(2 + 3)\n"
        "```\n"
        "Question: Which programming language is used? Answer with only the letter corresponding to the best answer.\n"
        "A) python\n"
        "B) java\n"
        "C) c++\n"
        "Answer:"
    )
    # fmt: on


def test_classication_render_few_shot_chat():
    problem = basic_classification_question_problem()
    few_shot_config = FewShotConfig(
        FewShotLibrary([basic_classification_question_problem_2()]),
        num_examples=1,
        change_examples_between_solves=False,
    )
    renderer = QuestionPromptRenderer()
    parser = CallbackResponseParser()
    renderer = FewShotRendererWrapper(
        renderer,
        parser,
        few_shot_config,
    )
    lm = MagicMock()
    lm.is_chat_model = True
    render = renderer.render(
        problem,
        lm,
    )
    print(render.prompt.text)
    # fmt: off
    assert render.prompt.is_text_a_chat()
    assert len(render.prompt.text) == 3
    assert render.prompt.text[0].content == (
        "```python\n"
        "if __name__ == '__main__':\n"
        "    print(2 + 3)\n"
        "```\n"
        "Question: Which value will be printed? Answer with only the letter corresponding to the best answer.\n"
        "A) 1\n"
        "B) 5\n"
        "C) hello world\n"
        "D) 2\n"
        "Answer:"
    )
    assert render.prompt.text[0].role == ChatGptRoles.user
    assert render.prompt.text[1].content == "B"
    assert render.prompt.text[1].role == ChatGptRoles.assistant
    assert render.prompt.text[2].content == (
        "```python\n"
        "if __name__ == '__main__':\n"
        "    print(2 + 3)\n"
        "```\n"
        "Question: Which programming language is used? Answer with only the letter corresponding to the best answer.\n"
        "A) python\n"
        "B) java\n"
        "C) c++\n"
        "Answer:"
    )
    assert render.prompt.text[2].role == ChatGptRoles.user
    # fmt: on


def run_test_simple_classification_solve(lm, known_ans: dict[str, float] | None = None):
    problem = basic_classification_question_problem()
    renderer = QuestionPromptRenderer()
    parser = CallbackResponseParser()
    solver = LmCodeSolverAutoRegressive(
        lm,
        prompt_renderer=renderer,
        response_parser=parser,
    )
    solution = solver.solve(problem)
    assert len(solution.solve_steps) == 1
    assert solution.solve_steps[0].path == "main.py"
    prob_dict = json.loads(solution.solve_steps[0].value)
    assert len(prob_dict) == 3
    assert set(prob_dict.keys()) == {"python", "java", "c++"}
    assert prob_dict["python"] > 0.5
    if known_ans is not None:
        assert len(prob_dict) == len(known_ans)
        for k, v in known_ans.items():
            assert math.isclose(prob_dict[k], v, abs_tol=0.0001)


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="No api in github actions")
def test_simple_classification_solve():
    lm = get_open_ai_lm()
    run_test_simple_classification_solve(lm)


@pytest.mark.skip(reason="need to merge hf top token")
def test_simple_classification_solve_hf():
    lm = get_huggingface_lm(
        "Salesforce/codegen-350M-multi",
        trust_remote_code=True,
    )
    run_test_simple_classification_solve(lm)


def test_simple_classification_solve_mocked():
    lm = get_open_ai_lm()
    mocked_pred = MagicMock()
    mocked_pred.completion_text = "A"
    mocked_pred.top_token_logprobs = [
        {
            "A": math.log(0.8),
        },
    ]
    lm.predict = MagicMock(return_value=mocked_pred)
    run_test_simple_classification_solve(
        lm,
        known_ans={"c++": 0.1, "python": 0.8, "java": 0.1},
    )


def test_simple_classification_solve_mocked2():
    lm = get_open_ai_lm()
    mocked_pred = MagicMock()
    mocked_pred.completion_text = "A"
    mocked_pred.top_token_logprobs = [
        {
            " a": math.log(0.8),
        },
    ]
    lm.predict = MagicMock(return_value=mocked_pred)
    run_test_simple_classification_solve(
        lm,
        known_ans={"c++": 0.1, "python": 0.8, "java": 0.1},
    )


def test_simple_classification_solve_mocked3():
    lm = get_open_ai_lm()
    mocked_pred = MagicMock()
    mocked_pred.completion_text = "A"
    mocked_pred.top_token_logprobs = [
        {
            "A": math.log(0.8),
            "B": math.log(0.15),
        },
    ]
    lm.predict = MagicMock(return_value=mocked_pred)
    run_test_simple_classification_solve(
        lm,
        known_ans={"c++": 0.05, "python": 0.8, "java": 0.15},
    )


def test_simple_classification_solve_mocked_norm_mass():
    lm = get_open_ai_lm()
    mocked_pred = MagicMock()
    mocked_pred.completion_text = "A"
    mocked_pred.top_token_logprobs = [
        {
            "A": math.log(0.4),
            "B": math.log(0.1),
            "C": math.log(0.00000000001),
        },
    ]
    lm.predict = MagicMock(return_value=mocked_pred)
    run_test_simple_classification_solve(
        lm,
        known_ans={"c++": 0.0, "python": 0.8, "java": 0.2},
    )


def test_simple_classification_agg():
    lm = get_open_ai_lm()
    mocked_pred = MagicMock()
    mocked_pred.completion_text = "A"
    mocked_pred.top_token_logprobs = [
        {
            "A": math.log(0.5),
            "B": math.log(0.1),
            "C": math.log(0.00000000001),
            " python": math.log(0.4),
            "sdf": math.log(0.4),
        },
    ]
    lm.predict = MagicMock(return_value=mocked_pred)
    run_test_simple_classification_solve(
        lm,
        known_ans={"c++": 0.0, "python": 0.9, "java": 0.1},
    )


def test_ans_to_solution():
    problem = basic_classification_question_problem()
    renderer = QuestionPromptRenderer(randomize_multiple_choice=False)
    parser = CallbackResponseParser()
    render = renderer.render(problem, lm=MagicMock())
    ans = parser.answer_for_solution(
        render,
        problem,
        solve_steps=[
            SolveStep(
                "main.py",
                problem.transformation_spec.statements[0].static_mark_id,
                "python",
            ),
        ],
    )
    assert ans == " A"
    ans = parser.answer_for_solution(
        render,
        problem,
        solve_steps=[
            SolveStep(
                "main.py",
                problem.transformation_spec.statements[0].static_mark_id,
                "java",
            ),
        ],
    )
    assert ans == " B"



from tree_sitter import Node

from synthegrator.util import get_first_line_with_nonwhitespace


def truncate_when_decreases_indent(text: str) -> str:
    first_line = get_first_line_with_nonwhitespace(text)
    starting_indent = len(first_line) - len(first_line.lstrip())
    lines = text.splitlines()
    seen_first_line = False
    for i, line in enumerate(lines):
        if line.lstrip() != "":
            seen_first_line = True
        if (
            len(line) - len(line.lstrip()) < starting_indent
            and seen_first_line
            and line.lstrip() != ""
        ):
            return "\n".join(lines[:i])
    return text


class TreeSitterUtil:
    @staticmethod
    def nodes_are_equal(n1: Node, n2: Node) -> bool:
        if n1 is None or n2 is None:
            return False
        return (
            n1.type == n2.type
            and n1.start_point == n2.start_point
            and n1.end_point == n2.end_point
        )

    @staticmethod
    def node_parent(tree, node) -> Node:
        to_visit = [tree.root_node]
        while len(to_visit) > 0:
            next_node = to_visit.pop()
            for child in next_node.children:
                if TreeSitterUtil.nodes_are_equal(child, node):
                    return next_node
            to_visit.extend(next_node.children)
        msg = "Could not find node in tree."
        raise ValueError(msg)

    @staticmethod
    def copy_tree_excluding_node(node, skip: Node, source_code: str):
        return TreeSitterUtil.copy_tree_excluding_nodes(node, [skip], source_code)

    @staticmethod
    def copy_tree_excluding_nodes(
        node: Node,
        skip: list[Node],
        source_code: str,
    ):
        new_source_code = ""

        # Iterate through the children and recursively call the function.
        child_index = 0
        skipped = False
        for child in node.children:
            if any(TreeSitterUtil.nodes_are_equal(child, _skip) for _skip in skip):
                skipped = True
            else:
                # Add the source code between this child and the previous child (or the start of the parent).
                new_source_code += source_code[
                    node.start_byte + child_index : child.start_byte
                ]

                # Recursively handle the child.
                new_source_code += TreeSitterUtil.copy_tree_excluding_nodes(
                    child,
                    skip,
                    source_code,
                )

            # Update the index to the end of this child.
            child_index = child.end_byte - node.start_byte

        # Add the source code after the last child.
        if skipped:
            new_source_code = new_source_code.lstrip()

        new_source_code += source_code[node.start_byte + child_index : node.end_byte]

        return new_source_code



import os

from aiohttp.pytest_plugin import pytest_generate_tests

from synthegrator.problem_rendering import render_single_edit_codeproblem_autoregressive
from synthegrator.synthdatasets.apps import yield_apps
import pytest

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


@pytest.mark.skip(
    reason="This is complaining about remote code. I could probably fix this "
           "but for now we aren't really using this dataset anyways."
)
def test_apps():
    first_apps_example = next(yield_apps(stream_subset=True))
    from pprint import pprint

    expected_text = (
        "An accordion is a string (yes, in the real world accordions are "
        "musical instruments, but let's forget about it for a"
    )
    out = render_single_edit_codeproblem_autoregressive(first_apps_example)
    assert expected_text in out.prompt.text
    assert out.changed_element is not None
    pprint(first_apps_example)


@pytest.mark.skip(
    reason="This is complaining about remote code. I could probably fix this "
           "but for now we aren't really using this dataset anyways."
)
def test_apps_2():
    val = list(yield_appstustream_subset=True)[5]
    assert render_single_edit_codeproblem_autoregressive(val)



#!/usr/bin/env python
import pandas as pd
from collections import Counter
from itertools import islice
from tqdm import tqdm

from lmwrapper.openai_wrapper import get_open_ai_lm, OpenAiModelNames

from synthegrator.code_solver import LmCodeSolverAutoRegressive
from synthegrator.df_converters import solution_evals_to_df
from synthegrator.execution_threading import solve_and_evaluate_problems
from synthegrator.synthdatasets.livecodebench import yield_livecode_problems
from synthegrator.synthdatasets.livecodebench.authorcode.benchmarks.code_generation import TestType, \
    Difficulty, load_code_generation_dataset


def count_test_types():
    data = load_code_generation_dataset()
    test_count = Counter()
    difficulty_count = Counter()
    for ex in tqdm(data):
        all_tcs = ex.private_test_cases + ex.public_test_cases
        assert all(all_tcs[0].testtype == tc.testtype for tc in all_tcs)
        test_count[all_tcs[0].testtype] += 1
        if all_tcs[0].testtype == TestType.FUNCTIONAL:
            #print("---")
            #print(ex.starter_code)
            pass
        difficulty_count[ex.difficulty] += 1
    print(test_count)
    print(difficulty_count)



def small_eval_live_code_bench():
    examples = islice(yield_livecode_problems(
        #filter_difficulty=Difficulty.EASY,
    ), 2000)
    #examples = list(examples)
    #print(len(examples))
    solver = LmCodeSolverAutoRegressive(
        get_open_ai_lm(
            OpenAiModelNames.gpt_4o,
            cache_outputs_default=True
        ),
    )
    evals = solve_and_evaluate_problems(
        solver,
        examples,
        max_threads_solve=3,
        max_threads_eval=1,
        finish_all_solves_first=True,
    )
    df = solution_evals_to_df(
        evals,
    )
    print(df)
    print("Solve rate")
    print(df.main_metric__is_success.mean())
    df.to_parquet("/tmp/livecodebench_evals.parquet")


def explore_solve_vals():
    df = pd.read_parquet("/tmp/livecodebench_evals.parquet")
    weird_main_count = 0
    weird_main_in_resp = 0
    for row in df.itertuples():
        if row.main_metric__is_success:
            continue
        print("---")
        solution_text = row.solution__solve_steps[0]['value']
        prompt = row.solution__prompt__text[0]['content']
        has_main_func = "def main(" in prompt
        has_input_call = "input(" in prompt or "sys.stdin" in prompt
        has_main_func_in_resp = "def main(" in solution_text
        print("main func", has_main_func)
        if has_main_func and not has_input_call:
            #print(prompt)
            weird_main_count += 1
        if has_main_func_in_resp:
            weird_main_in_resp += 1
            print(solution_text)
        #print(prompt)
    print(f"weird main count {weird_main_count}/{len(df)} = {weird_main_count/len(df)}")
    print(f"weird main in resp {weird_main_in_resp}/{len(df)} = {weird_main_in_resp/len(df)}")
    print("solve rate", df.main_metric__is_success.mean())


if __name__ == "__main__":
    count_test_types()
    exit()
    small_eval_live_code_bench()
    explore_solve_vals()


from unittest.mock import MagicMock

import pytest

from synthegrator.code_problem_builders import (
    make_simple_method_completion_problem,
    make_solve_for_single_edit,
)
from synthegrator.code_problems import (
    CodeProblem,
    CodeSolution,
    DiscoveredTestsuite,
    TestCaseMethodCallIsEq,
)
from synthegrator.environments import ProjectWorkingDirectory
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec
from synthegrator.memory_fs import ProjectDir
from synthegrator.problem_rendering import render_single_edit_codeproblem_autoregressive
from synthegrator.solution_eval import (
    TestRunResultList,
    apply_solution,
    apply_steps_to_markup,
    create_exact_match_metric,
    evaluate_code_problem_execution,
    parse_junit_test_cases,
)
from synthegrator.tests.non_dataset_tests.test_value_question_render import (
    basic_classification_question_problem,
)
from synthegrator.transformation_spec import (
    MarkElement,
    MarkText,
    SolveStep,
    StsInsert,
    StsInvisible,
    StsNormalText,
    StsSelectorCharSpan,
    StsSelectorLineSpan,
    StsSpecStatement,
    TransformationSpec,
)
from synthegrator.util import get_only_item


def test_evaluate_code_problem():
    problem = make_simple_method_completion_problem(
        signature="def foo(x) -> int:",
        prompt="Write a function that doubles its input.",
        test_cases=[
            TestCaseMethodCallIsEq(
                test_input=(2,),
                output=4,
                method_name="foo",
                is_hidden_test=False,
                fail_message="",
            ),
            TestCaseMethodCallIsEq(
                test_input=(3,),
                output=6,
                method_name="foo",
                is_hidden_test=False,
                fail_message="",
            ),
        ],
    )
    solve = make_solve_for_single_edit(problem, "    return x * 2")
    result_list = evaluate_code_problem_execution(
        problem,
        solve,
        user_prompted_guard=False,
    )
    assert result_list[0].success
    assert result_list[1].success
    assert result_list.all_passed()
    solve_eval = problem.preferred_solution_evaluator.evaluate(solve)
    assert solve_eval.main_metric.is_success
    assert solve_eval.main_metric.float_val == 1.0
    assert solve_eval.main_metric.label == TestRunResultList.METRIC_NAME


def test_evaluate_code_problem_wrong():
    problem = make_simple_method_completion_problem(
        signature="def foo(x) -> int:",
        prompt="Write a function that doubles its input.",
        test_cases=[
            TestCaseMethodCallIsEq(
                test_input=(2,),
                output=4,
                method_name="foo",
                is_hidden_test=False,
                fail_message="",
            ),
            TestCaseMethodCallIsEq(
                test_input=(3,),
                output=6,
                method_name="foo",
                is_hidden_test=False,
                fail_message="",
            ),
        ],
    )
    solve = make_solve_for_single_edit(problem, "    return 4")
    result_list = evaluate_code_problem_execution(
        problem,
        solve,
        user_prompted_guard=False,
    )
    assert result_list[0].success
    assert not result_list[1].success
    assert not result_list.all_passed()
    solve_eval = problem.preferred_solution_evaluator.evaluate(solve)
    assert not solve_eval.main_metric.is_success
    assert solve_eval.main_metric.float_val == 0.5
    assert solve_eval.main_metric.label == TestRunResultList.METRIC_NAME


def test_apply_steps_to_markup():
    node = MarkElement(
        StsNormalText(),
        [
            MarkText("foo", 0, 3),
            MarkElement(StsInvisible(), [MarkText("bar", 3, 6)], 3, 6),
            MarkText("baz", 6, 9),
            MarkElement(StsInsert(), [], 9, 9),
        ],
        0,
        9,
    ).freeze()
    insert_node = node.children[3]
    assert isinstance(insert_node.verb, StsInsert)
    val = apply_steps_to_markup(node, [])
    assert val == "foobarbaz"
    val = apply_steps_to_markup(node, [SolveStep("foo.py", insert_node.mark_id, "qux")])
    assert val == "foobarbazqux"
    val = apply_steps_to_markup(node, [SolveStep("foo.py", "wrong", "qux")])
    assert val == "foobarbaz"
    val = apply_steps_to_markup(
        node,
        [SolveStep("foo.py", insert_node.mark_id, "qux")],
        still_hide_invisible=True,
    )
    assert val == "foobazqux"


def test_apply_steps_to_markup2():
    node = MarkElement(
        StsNormalText(),
        [
            MarkText("foo", 0, 3),
            MarkElement(StsInvisible(), [MarkText("bar", 3, 6)], 3, 6),
            MarkElement(StsInsert(), [], 6, 6),
            MarkText("baz", 6, 9),
        ],
        0,
        9,
    ).freeze()
    insert_node = node.children[2]
    assert isinstance(insert_node.verb, StsInsert)
    val = apply_steps_to_markup(node, [])
    assert val == "foobarbaz"
    val = apply_steps_to_markup(node, [SolveStep("foo.py", insert_node.mark_id, "qux")])
    assert val == "foobarquxbaz"
    val = apply_steps_to_markup(node, [SolveStep("foo.py", "wrong", "qux")])
    assert val == "foobarbaz"
    val = apply_steps_to_markup(
        node,
        [SolveStep("foo.py", insert_node.mark_id, "qux")],
        stop_after_first_change=True,
    )
    assert val == "foobarqux"


def test_fail_module():
    # reference: https://github.com/windyroad/JUnit-Schema/blob/master/JUnit.xsd#L73C7-L73C16
    xml = """<?xml version="1.0" encoding="utf-8"?>
<testsuites>
    <testsuite name="pytest" errors="1" failures="0" skipped="0" tests="1" time="0.044" timestamp="2000-02-27T00:53:12.379941" hostname="530000000a3d">
        <testcase classname="" name="synth_solution" time="0.000">
            <error message="collection failure">ImportError while importing test module
                '/sandbox/synth_solution.py'.
                Hint: make sure your test modules/packages have valid Python names.
                Traceback:
                /usr/local/lib/python3.11/importlib/__init__.py:126: in import_module
                return _bootstrap._gcd_import(name[level:], package, level)
                synth_solution.py:2: in &lt;module&gt;
                from interp.nlptools_custom.nlp_utils import get_wiki_entities, get_wiki_element,
                find_entailment_ratings, get_nlp
                E ModuleNotFoundError: No module named 'interp'
            </error>
        </testcase>
        <testcase classname="" name="synth_solution1" time="0.000">
            <failure message="AssertionError: assert 1 == 0">
    def test_new_sample_no_eval(example_dypy_prob):
        problem_w_doc, function_w_doc = example_dypy_prob
        solver = DummyCodeSolverAutoRegressive(None)
        solution = solver.solve(problem_w_doc, lang_spec=PythonLangSpec())
&gt;       assert len(solution.solve_steps) == 0
E       AssertionError: assert 1 == 0
E        +  where 1 = len((SolveStep(path='project1/grab/document.py', mark_id='tag:StsEditable,4099,4161,1', value='            return unicodec.normalize_encoding_name(encoding)\n'),))
E        +    where (SolveStep(path='project1/grab/document.py', mark_id='tag:StsEditable,4099,4161,1', value='            return unicodec.normalize_encoding_name(encoding)\n'),) = CodeSolution(solve_steps=(SolveStep(path='project1/grab/document.py', mark_id='tag:StsEditable,4099,4161,1', value='  ... instructions_are_essential=False), lang_spec=&lt;synthegrator.lang_spec.PythonLangSpec object at 0x122340f50&gt;, time=None).solve_steps

synthegrator/tests/test_xx_dypybench.py:177: AssertionError
            </failure>
        </testcase>
        <testcase classname="" name="synth_solution2" time="0.000">
            <failure></failure>
        </testcase>
        <testcase classname="" name="synth_solution3" time="0.000">
        </testcase>
    </testsuite>
</testsuites>""".strip()
    results = parse_junit_test_cases(xml, MagicMock())
    assert results.summary.valid_xml
    assert len(results) == 4
    assert not results.all_passed()
    assert not results[0].success
    assert "Hint: " in results[0].fail_text
    assert results[0].fail_message == "collection failure"

    assert not results[1].success
    assert "def test_new_sample_no_eval" in results[1].fail_text
    assert "AssertionError" in results[1].fail_message


def test_invalid_result_xml():
    # reference: https://github.com/windyroad/JUnit-Schema/blob/master/JUnit.xsd#L73C7-L73C16
    xml = """<?xml version="1.0" encoding="utf-8"?>
<testsuites>
    <testsuite name="pytest" errors="1" failures="0" skipped="0" tests="1" time="0.044" timestamp="2000-02-27T00:53:12.379941" hostname="530000000a3d">
        <testcase classname="" name="synth_solution" time="0.000">
            <error message="collection failure">ImportError while importing test module
                '/sandbox/synth_solution.py'.
                Hint: make sure your test modules/packages have valid Python names.
                Traceback:
                /usr/local/lib/python3.11/importlib/__init__.py:126: in import_module
                return _bootstrap._gcd_import(name[level:], package, level)
                synth_solution.py:2: in &lt;module&gt;
                from interp.nlptools_custom.nlp_utils import get_wiki_entities, get_wiki_element,
                find_entailment_ratings, get_nlp
                E ModuleNotFoundError: No module named 'interp'
            </error>
            <failure message="AssertionError: assert 1 == 0">
    def test_new_sample_no_eval(example_dypy_prob):
        problem_w_doc, function_w_doc = example_dypy_prob
        solver = DummyCodeSolverAutoRegressive(None)
        solution = solver.solve(problem_w_doc, lang_spec=PythonLangSpec())
&gt;       assert len(solution.solve_steps) == 0
E       AssertionError: assert 1 == 0
E        +  where 1 = len((SolveStep(path='project1/grab/document.py', mark_id='tag:StsEditable,4099,4161,1', value='            return unicodec.normalize_encoding_name(encoding)\n'),))
E        +    where (SolveStep(path='project1/grab/document.py', mark_id='tag:StsEditable,4099,4161,1', value='            return unicodec.normalize_encoding_name(encoding)\n'),) = CodeSolution(solve_steps=(SolveStep(path='project1/grab/document.py', mark_id='tag:StsEditable,4099,4161,1', value='  ... instructions_are_essential=False), lang_spec=&lt;synthegrator.lang_spec.PythonLangSpec object at 0x122340f50&gt;, time=None).solve_steps

synthegrator/tests/test_xx_dypybench.py:177: AssertionError
            </failure>
        </testcase>
        <testcase classname="" name="synth_solution2" time="0.000">
            <failure></failure>
        </testcase>
        <testcase classname="" name="synth_solution3" time="0.000">
        </testcase>
    </testsuite>
</testsuites>""".strip()
    results = parse_junit_test_cases(xml, MagicMock())
    assert not results.summary.valid_xml
    assert len(results) == 3
    assert not results.all_passed()
    assert not results[0].success
    # This defacto means the last failure/error message will be used!
    assert "def test_new_sample_no_eval" in results[0].fail_text
    assert "AssertionError" in results[0].fail_message

    assert not results[1].success
    assert results[1].fail_text is None
    assert results[1].fail_message is None


def test_discovered_test_cases():
    test_case = DiscoveredTestsuite(
        cmds=[
            # Write some "existing" test suite to a file
            (
                "echo 'import example\ndef test_foo():\n    assert example.foo() == 2'"
                " > /sandbox/test_example.py"
            ),
            # Run pytest on it and get the result
            "python3 -m pytest --junitxml=/sandbox/junit.xml /sandbox/test_example.py",
            "cat /sandbox/junit.xml",
        ],
    )
    problem = CodeProblem(
        working_directory=ProjectWorkingDirectory(
            files=ProjectDir.construct_with_one_file(
                "example.py",
                "def foo():\n    return None",
            ),
        ),
        transformation_spec=TransformationSpec(
            statements=(
                StsSpecStatement(None, StsSelectorLineSpan(1, 2), StsInsert()),
            ),
        ),
        test_cases=[test_case],
    )
    solution = make_solve_for_single_edit(problem, "    return 2")
    result_list = evaluate_code_problem_execution(problem, solution)
    assert len(result_list) == 1
    assert result_list[0].completed
    assert result_list[0].success
    assert result_list.all_passed()


def test_exception_on_bad_solution_indent():
    body = "def foo():\n"
    problem = CodeProblem(
        working_directory=ProjectWorkingDirectory(
            files=ProjectDir.construct_with_one_file("example.py", body),
        ),
        transformation_spec=TransformationSpec(
            statements=(
                StsSpecStatement(
                    None,
                    StsSelectorCharSpan(len(body), len(body)),
                    StsInsert(
                        set_base_space_indent=4,
                    ),
                ),
            ),
        ),
    )
    render = render_single_edit_codeproblem_autoregressive(problem)
    solution = CodeSolution(
        solve_steps=(
            SolveStep("example.py", render.changed_element.mark_id, "    return 2"),
        ),
        problem=problem,
        lang_spec=PythonLangSpec(),
    )
    sol = apply_solution(solution)
    assert sol.get_only_file().content_str == "def foo():\n    return 2"
    solution = CodeSolution(
        solve_steps=(
            SolveStep("example.py", render.changed_element.mark_id, "return 2"),
        ),
        problem=problem,
        lang_spec=PythonLangSpec(),
    )
    with pytest.raises(ValueError):
        apply_solution(solution)


def test_classification_question_eval():
    problem = basic_classification_question_problem()
    classification = get_only_item(
        problem.transformation_spec.classification_question_statements(),
    )
    solution = CodeSolution(
        solve_steps=(
            SolveStep(
                classification.attempt_to_get_path_as_exact_path(),
                classification.static_mark_id,
                "python",
            ),
        ),
        problem=problem,
        lang_spec=PythonLangSpec(),
    )
    problem.known_solutions.append(solution)
    eval = problem.preferred_solution_evaluator.evaluate(solution)
    assert eval.main_metric.is_success
    assert len(eval.extra_metrics) == 3
    assert eval.get_metric_from_key("classification_correct_prob").float_val == 1.0
    solution_2 = CodeSolution(
        solve_steps=(
            SolveStep(
                classification.attempt_to_get_path_as_exact_path(),
                classification.static_mark_id,
                '{"python": 0.1, "java": 0.2, "c++": 0.7}',
            ),
        ),
        problem=problem,
        lang_spec=PythonLangSpec(),
    )
    eval = problem.preferred_solution_evaluator.evaluate(solution_2)
    assert not eval.main_metric.is_success
    assert eval.get_metric_from_key("classification_correct_prob").float_val == 0.1
    print(eval)


def test_exact_match():
    problem = make_simple_method_completion_problem(
        signature="def foo(x) -> int:",
        prompt="Write a function that doubles its input.",
        known_solutions=["    return 4"],
        test_cases=[
            TestCaseMethodCallIsEq(
                test_input=(2,),
                output=4,
                method_name="foo",
                is_hidden_test=False,
                fail_message="",
            ),
            TestCaseMethodCallIsEq(
                test_input=(3,),
                output=6,
                method_name="foo",
                is_hidden_test=False,
                fail_message="",
            ),
        ],
    )
    solve = make_solve_for_single_edit(problem, "    return 4")

    evaluator = problem.preferred_solution_evaluator
    evaluator._extra_metric_funcs = [create_exact_match_metric(strip=True)]

    solve_eval = evaluator.evaluate(solve)

    # one of the actual tests should pass
    assert not solve_eval.main_metric.is_success
    assert solve_eval.main_metric.float_val == 0.5
    assert solve_eval.main_metric.label == TestRunResultList.METRIC_NAME

    # and we expect an exact match
    assert solve_eval.extra_metrics[0].is_success
    assert solve_eval.extra_metrics[0].float_val == 1.0
    assert solve_eval.extra_metrics[0].label == "exact_match"

    # The tests should fail but exact match should succeed
    # due to strip
    solve = make_solve_for_single_edit(problem, "     return 4  ")

    evaluator = problem.preferred_solution_evaluator
    evaluator._extra_metric_funcs = [create_exact_match_metric(strip=True)]

    solve_eval = evaluator.evaluate(solve)

    # the actual tests should pass
    assert not solve_eval.main_metric.is_success
    assert solve_eval.main_metric.float_val == 0.0
    assert solve_eval.main_metric.label == TestRunResultList.METRIC_NAME

    # and we expect an exact match
    assert solve_eval.extra_metrics[0].is_success
    assert solve_eval.extra_metrics[0].float_val == 1.0
    assert solve_eval.extra_metrics[0].label == "exact_match"

    # The tests should pass but exact match should fail
    solve = make_solve_for_single_edit(problem, "    return x*2")

    evaluator = problem.preferred_solution_evaluator
    evaluator._extra_metric_funcs = [create_exact_match_metric(strip=True)]

    solve_eval = evaluator.evaluate(solve)

    # the actual tests should pass
    assert solve_eval.main_metric.is_success
    assert solve_eval.main_metric.float_val == 1.0
    assert solve_eval.main_metric.label == TestRunResultList.METRIC_NAME

    # and we expect an exact match
    assert not solve_eval.extra_metrics[0].is_success
    assert solve_eval.extra_metrics[0].float_val == 0.0
    assert solve_eval.extra_metrics[0].label == "exact_match"

    # Half the tests should pass but exact match should fail
    solve = make_solve_for_single_edit(problem, "    return 4 ")

    evaluator = problem.preferred_solution_evaluator
    evaluator._extra_metric_funcs = [create_exact_match_metric(strip=False)]

    solve_eval = evaluator.evaluate(solve)

    # the actual tests should pass
    assert not solve_eval.main_metric.is_success
    assert solve_eval.main_metric.float_val == 0.5
    assert solve_eval.main_metric.label == TestRunResultList.METRIC_NAME

    # and we expect an exact match
    assert not solve_eval.extra_metrics[0].is_success
    assert solve_eval.extra_metrics[0].float_val == 0.0
    assert solve_eval.extra_metrics[0].label == "exact_match"



import json
import logging
logger = logging.getLogger(__name__)
from collections.abc import Iterator
from typing import Any

from synthegrator.code_problem_builders import (
    make_simple_file_generation,
    make_simple_method_completion_problem,
)
from synthegrator.code_problems import (
    CodeProblem,
    TestCase,
    TestCaseArbitraryMethod,
)
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec

APPS_CODE_ENV = (
    "import sys\nimport time\nimport itertools\nfrom itertools import accumulate,"
    " product, permutations, combinations\nimport collections\nfrom collections import"
    " Counter, OrderedDict, deque, defaultdict, ChainMap\nfrom functools import"
    " lru_cache\nimport math\nfrom math import sqrt, sin, cos, tan, ceil, fabs, floor,"
    " gcd, exp, log, log2\nimport fractions\nfrom typing import List, Tuple\nimport"
    " numpy as np\nimport random\nimport heapq\nfrom heapq import *\n"
)


def produce_app_file_generation(
    task_id: str,
    prompt: str,
    canonical_solutions: list[str],
    inputs_outputs: dict[str, list[str]],
    starter_code: str,
) -> None:
    assert not isinstance(canonical_solutions, str), "Expect list of solutions?"
    return make_simple_file_generation(
        prompt=prompt,
        test_cases=parse_input_output_to_test_case(inputs_outputs, starter_code),
        model_context="",
        execution_context_hidden_from_model=APPS_CODE_ENV,
        dataset_name="apps",
        problem_id=task_id,
        known_solutions=canonical_solutions,
    )


def produce_app_method_problem(
    task_id: str,
    prompt: str,
    canonical_solutions: list[str],
    inputs_outputs: dict[str, list[str]],
    starter_code: str,
) -> CodeProblem:
    signature = starter_code.rstrip().replace("\t", "    ") if starter_code else ""
    assert not isinstance(canonical_solutions, str), "Expect list of solutions?"
    return make_simple_method_completion_problem(
        prompt,
        signature=signature,
        test_cases=parse_input_output_to_test_case(inputs_outputs, starter_code),
        problem_id=task_id,
        known_solutions=canonical_solutions,
        lang_spec=PythonLangSpec(),
        execution_context_hidden_from_model=APPS_CODE_ENV,
        prompt_already_docstring=True,
    )


def yield_apps(stream_subset=False) -> Iterator[CodeProblem]:
    from datasets import load_dataset

    # for ci, take 10 examples only
    if stream_subset:
        dataset = load_dataset(
            "codeparrot/apps",
            split="test",
            streaming=True,  # , trust_remote_code=True
        ).take(10)
    else:
        dataset = load_dataset(
            "codeparrot/apps",
            split="test",  # , trust_remote_code=True
        ).shuffle()

    for ex in dataset:
        # if no example input/output and no available solutions, skip
        if not (ex["input_output"] and ex["solutions"]):
            continue

        # load solutions
        try:
            ex["solutions"] = json.loads(ex["solutions"])  # TODO only one solution?
        except ValueError:
            logger.info("Unable to load solution %s", ex["solutions"])
            continue

        try:
            ex["input_output"] = json.loads(ex["input_output"])
        except ValueError:
            # skip if unable to load input/output example
            continue

        # see apps/eval/testing_util.py:145
        if ex["input_output"].get("fn_name") is None:
            # function has no name -> module level
            yield produce_app_file_generation(
                task_id=ex["problem_id"],
                prompt=ex["question"],
                canonical_solutions=ex["solutions"],
                inputs_outputs=ex["input_output"],
                starter_code=ex["starter_code"],
            )
        else:
            # function has a name
            yield produce_app_method_problem(
                task_id=ex["problem_id"],
                prompt=ex["question"],
                canonical_solutions=ex["solutions"],
                inputs_outputs=ex["input_output"],
                starter_code=ex["starter_code"],
            )


def parse_input_output_to_test_case(
    inputs_outputs: dict[str, any],
    starter_code: str,
) -> list[TestCase]:
    # TODO split several asserts into different test cases
    # pull out just the body
    inputs = inputs_outputs["inputs"]
    outputs = inputs_outputs["outputs"]
    fn_name = inputs_outputs.get("fn_name")

    test_cases = []
    for i, input_ex in enumerate(inputs):
        # see apps/eval/testing_util.py:243
        # JSON forces dictionaries to have string keys; this undoes this (assuming a singleton list)
        try:
            if isinstance(inputs[0], dict):
                inputs = [{int(k): v for k, v in inputs[0].items()}]
        except KeyError:
            pass

        try:
            if isinstance(outputs[i], dict):
                outputs[i] = [{int(k): v for k, v in outputs[i].items()}]
        except KeyError:
            pass

        try:
            if isinstance(outputs[i][0], dict):
                outputs[i] = [{int(k): v for k, v in outputs[i][0].items()}]
        except KeyError:
            pass

        output_ex = outputs[i]

        if isinstance(input_ex, str) and isinstance(output_ex, str) and not fn_name:
            code = _test_case_for_builtins_str(
                i,
                input_ex.replace("\\", "\\\\"),
                output_ex.replace("\\", "\\\\"),
            )
        elif not isinstance(input_ex, str) and not fn_name:
            code = _test_case_for_builtins(i, input_ex, output_ex)
        elif fn_name:
            code = _test_case_for_named_function(
                i,
                input_ex,
                output_ex,
                fn_name,
                starter_code,
            )

        test_case = TestCaseArbitraryMethod(
            is_hidden_test=True,
            fail_message=None,
            method_body=code,
            lang_spec=PythonLangSpec(),
            fixtures=("monkeypatch", "capfd"),
        )
        test_cases.append(test_case)
    return test_cases


def _test_case_for_builtins_str(i: int, function_input: str, output: str) -> str:
    return f"""
inputs = iter(\"\"\"{function_input}\"\"\".split("\\n"))
monkeypatch.setattr('builtins.input', lambda: next(inputs))
main()
check_outputs(capfd)
out, err = capfd.readouterr()
assert out == \"\"\"{output}\"\"\"
""".strip()


def _test_case_for_builtins(i: int, function_input: any, output: any) -> str:
    return f"""
inputs = iter({function_input})
monkeypatch.setattr('builtins.input', lambda: next(inputs))
main()
out, err = capfd.readouterr()
assert out == {output}
""".strip()


def apps_assertion_code_gen(x: Any) -> str:
    match str(type(x).__name__):
        case "str":
            y = '["' + x + '"]'
        case "list":
            y = str(x)
        case "bool":
            y = str(x)
        case "int":
            y = str(x)
        case _:
            msg = "Unsupported type"
            raise Exception(msg, type(x))
    return y


def _test_case_for_named_function(
    i: int,
    ref_input: Any,
    ref_output: Any,
    fn_name: str,
    starter_code: str,
) -> str:
    args = apps_assertion_code_gen(ref_input)
    out = apps_assertion_code_gen(ref_output)

    """
    TODO: the reference implementation works like this:
        if isinstance(ref_output, tuple):
            output = list(output)

        tmp_result = output == ref_output
        if isinstance(ref_input, list) and ref_output:
            tmp_result = tmp_result or (output == ref_input[0])

        # ground truth sequences are not tuples
        try:
            if isinstance(output[0], tuple):
                tmp_result = tmp_result or ([list(x) for x in output] == output[0])
        except:
            True
    """

    if "class Solution:" in starter_code:  # TODO replace with regex or treesitter
        if f"{fn_name}(self," in starter_code:
            code = f"""assert Solution().{fn_name}(*{args}) == {out}"""
        else:
            code = f"""assert Solution.{args}) == {out}"""
    else:
        code = f"""assert {fn_name}({args}) == {out}"""
    return code



from pathlib import Path
from lmwrapper.openai_wrapper import get_open_ai_lm, OpenAiModelNames
import pandas as pd

cur_path = Path(__file__).parent


def _make_df(max_problems, model_name, solver_type, cache = True):
    key = f"{model_name}_{solver_type}_{max_problems}"
    print(key)
    save_path = cur_path / f"repocod_evals_{key}.parquet"
    if cache and save_path.exists():
        return pd.read_parquet(save_path), key
    raise NotImplementedError("Not implemented")


if __name__ == "__main__":
    pass_fracs = {}
    max_problems = 1000
    all_dfs = []
    for model_name, solver_type in [
        (OpenAiModelNames.gpt_4_1, "dummy"),
        #(OpenAiModelNames.gpt_4_1, "autoregressive"),
        (OpenAiModelNames.gpt_4_1_mini, "autoregressive"),
        #(OpenAiModelNames.gpt_4_1, "tagged"),
        (OpenAiModelNames.gpt_4_1_mini, "tagged"),
        (OpenAiModelNames.gpt_4o_mini, "autoregressive"),
    ]:
        df, key = _make_df(max_problems, model_name, solver_type)
        pass_frac = df.main_metric__is_success.mean()
        print(f"Fraction Passing {key}: {pass_frac}")
        pass_fracs[key] = pass_frac
        df["key"] = key
        def get_repo_name(problem_id):
            r = problem_id.split("/")[0]
            r = r.split("_")[1]
            return r
        df["repo_name"] = df.solution__problem__problem_id.apply(get_repo_name)
        all_dfs.append(df)

    all_df = pd.concat(all_dfs)
    print(all_df)


from synthegrator.util import MultilineString

FIXTURE = """# shebang
import pandas

class Hello():
    ...

def method(blah):
    # hi
    x = 1+2
    y = x+3

    return x+y
"""


def test_line_range():
    string = MultilineString(FIXTURE)
    assert (
        string[9:]
        == """    y = x+3

    return x+y
"""
    )


def test_str_eq():
    string = MultilineString("test")
    assert string == "test"


def test_str_rep():
    string = MultilineString("test")
    assert str(string) == "test"


def test_str():
    string = MultilineString(FIXTURE)
    expected = [
        "# shebang\n",  # 1
        "import pandas\n",  # 2
        "\n",  # 3
        "class Hello():\n",  # 4
        "    ...\n",  # 5
        "\n",  # 6
        "def method(blah):\n",  # 7
        "    # hi\n",  # 8
        "    x = 1+2\n",  # 9
        "    y = x+3\n",  # 10
        "\n",  # 11
        "    return x+y\n",  # 12
    ]
    assert string.lines == expected
    assert string.line_count == 12
    assert len(string) == 12
    new_code = """    x = 20
"""
    string[8] = new_code
    expected = """# shebang
import pandas

class Hello():
    ...

def method(blah):
    # hi
    x = 20
    y = x+3

    return x+y
"""
    assert string == expected


def test_insert():
    string = MultilineString(FIXTURE)

    new_code = """    z = 30
"""
    expected = """# shebang
import pandas

class Hello():
    ...

def method(blah):
    # hi
    x = 1+2
    z = 30
    y = x+3

    return x+y
"""
    string.insert(9, new_code)
    assert string == expected


def test_multiline_replace():
    string = MultilineString(FIXTURE)

    new_code = """    x = 20
    z = 1
    y = 5+x
"""
    string[8:11] = new_code
    expected = """# shebang
import pandas

class Hello():
    ...

def method(blah):
    # hi
    x = 20
    z = 1
    y = 5+x
    return x+y
"""
    assert string == expected



import difflib
import hashlib
import inspect
import string
from collections import UserString
from collections.abc import Callable, Iterable, Iterator, Sequence
from functools import wraps
from typing import SupportsIndex

import numpy as np


def ensure_newline_at_end(s: str) -> str:
    if s.endswith("\n"):
        return s
    return s + "\n"


def assert_str_eq_with_diff_error(a: str, b: str):
    if a != b:
        # hacky get the two expressions of the caller line
        # https://stackoverflow.com/a/50636091/382912
        import inspect

        caller_frame = inspect.stack()[1]
        caller_line = caller_frame.code_context[0].strip()
        caller_line = caller_line.split("assert_str_eq_with_diff_error(")[1]
        caller_line = caller_line.split(")")[0]
        a_name, b_name = caller_line.split(", ")
        diff = difflib.unified_diff(
            a.splitlines(),
            b.splitlines(),
            fromfile=a_name,
            tofile=b_name,
        )
        diff_str = "\n".join(diff)
        msg = f"Strings not equal:\n{diff_str}"
        raise AssertionError(msg)


def pretty_print_python_code(code: str):
    """Pretty prints python code with syntax highlighting."""
    import pygments
    from pygments.formatters import TerminalFormatter
    from pygments.lexers import PythonLexer

    print(pygments.highlight(code, PythonLexer(), TerminalFormatter()))


def n_lines(code: str, n: int) -> str:
    lines = code.split("\n")
    return "\n".join(lines[: min(n, len(lines))])


def clip(value, min_value, max_value):
    if value < min_value:
        return min_value
    if value > max_value:
        return max_value
    return value


class MultilineString(UserString):
    def __init__(self, string: str, expect_newline: bool = False) -> None:
        super().__init__(string)
        self.lines = string.splitlines(keepends=True)
        self.expect_newline = expect_newline

    def _new_line_guard(self, value: str) -> str:
        if self.expect_newline and not value.endswith("\n"):
            msg = "All strings must end in a newline!"
            raise Exception(msg)

        if not value.endswith("\n"):
            return value + "\n"
        return value

    @property
    def _string(self) -> str:
        return "".join(self.lines)

    def __str__(self) -> str:
        return self._string

    def __getitem__(self, index: SupportsIndex | slice) -> str:
        return "".join(self.lines[index])

    def __setitem__(self, index: SupportsIndex | slice, value: str) -> str:
        if not isinstance(value, str):
            msg = "Currently only a string is a valid value."
            raise TypeError(msg)
        value = self._new_line_guard(value)
        self.lines[index] = value
        self.data = self._string
        return self

    def __delitem__(self, index: SupportsIndex | slice) -> None:
        del self.lines[index]
        self.data = self._string

    def __len__(self) -> int:
        return len(self.lines)

    def insert(self, index: SupportsIndex | slice, lines: str | list) -> str:
        if isinstance(lines, str):
            lines = self._new_line_guard(lines)
            lines = lines.splitlines(keepends=True)

        self.lines[index:index] = lines
        self.data = self._string
        return self

    @property
    def line_count(self) -> int:
        return len(self.lines)


def lazy_prop(func):
    val = None

    def inner():
        nonlocal val
        if val is None:
            val = func()
        return val

    return inner


def estimate_best_overlap_span(
    larger_str: str,
    larger_tokens: list[str] | None,
    mod_string: str,
    tokenizer_func: Callable[[str], list[str]],
) -> tuple[int, int]:
    """
    Returns a slice into a sequence of larger tokens that might best overlap
    with smaller tokens. This is designed for when the smaller string is a
    chopped and slightly modified (such as adding preceeding indent) version
    """
    if larger_tokens is None:
        larger_tokens = tokenizer_func(larger_str)
    if larger_str == mod_string:
        return 0, len(larger_tokens)
    smaller_tokens = tokenizer_func(mod_string)
    large_t = np.array(larger_tokens)
    smaller_tokens = np.array(smaller_tokens)
    # TODO: I think an algorithm that looks like making a edit distance
    #   dp matrix first, then finding the offset as a path through that
    #   can work.
    for start_point in range(len(large_t) - len(smaller_tokens)):
        end_point = start_point + len(smaller_tokens)
        if np.array_equal(large_t[start_point:end_point], smaller_tokens):
            return start_point, end_point
    # For now just give up and return from beginning
    return 0, len(larger_tokens)


def frozen_method_cache(func: Callable) -> Callable:
    """
    A decorator for methods. It is expected this only on non-mutable clases.
    The `self` is ignored in the cache. If you use it on mutable classes that
    depend on the class state, then it might get unexpected results.
    """
    uncached_sentinel = object()  # Unique. Allows a return of None to be cached
    cache = uncached_sentinel

    if len(inspect.signature(func).parameters) == 1:
        uncached_sentinel = object()  # Unique. Allows a return of None to be cached
        cache = uncached_sentinel

        @wraps(func)
        def wrapper(self):
            nonlocal cache
            if cache is uncached_sentinel:
                cache = func(self)
            return cache

    else:
        cache = {}

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            key = (args, tuple(kwargs.items()))
            if key not in cache:
                cache[key] = func(self, *args, **kwargs)
            return cache[key]

    return wrapper


def get_first_line_with_nonwhitespace(value: str) -> str:
    try:
        first_line = next(
            line for line in value.splitlines(keepends=True) if line.strip() != ""
        )
    except StopIteration:
        first_line = value
    return first_line


from typing import TypeVar

T = TypeVar("T")


class IteratorWithLength(Iterator[T]):
    def __init__(self, iterable: Iterator[T], length: int | None) -> None:
        self.iterable = iterable
        self.length = length
        self.num_yielded = 0

    def __iter__(self) -> Iterator[T]:
        return self

    def __next__(self) -> T:
        if self.length is not None and self.num_yielded >= self.length:
            raise StopIteration
        self.num_yielded += 1
        return next(self.iterable)

    def __len__(self) -> int | None:
        if self.length is None:
            return None
        return self.length - self.num_yielded


# Characters used for Base62 encoding
BASE62 = string.digits + string.ascii_letters


def base62_encode(num):
    if num == 0:
        return BASE62[0]
    base62 = []
    while num:
        num, rem = divmod(num, 62)
        base62.append(BASE62[rem])
    return "".join(reversed(base62))


def hash_string_base62(
    s: str,
    digest_size_bytes: int = 8,
) -> str:
    # Using blake2b for a faster, non-cryptographic hash.
    # Adjust the digest_size if you want a shorter output.
    hash_bytes = hashlib.blake2b(s.encode(), digest_size=digest_size_bytes).digest()
    # Convert hash bytes to an integer
    num = int.from_bytes(hash_bytes, byteorder="big")
    # Encode this number in base62
    return base62_encode(num)


def find_after_first_occurrence_of_string(
    string_parts: list[str],
    split_str: str,
) -> tuple[int, int] | tuple[None, None]:
    """
    We are given `string_parts` list, which represents a
    string that is a concatenation of the strings in the list.
    That is, if you were to join all the parts, you would get one
    string. We are searching for the index that begins a string
    after the `split_str`. Because it might split cleanly within
    string, we also return the index within the string part itself
    which begins after the split str

    If `search_str` is not found, returns (None, None)

    Examples
    --------
        >> find_after_first_occurrence_of_string(
            ["My", " Name", " Is"],
            "Name"
        ) == (2, 0)
        # The split string might span between parts
        >> find_after_first_occurrence_of_string(
            ["My", " Name", " Is", " Bob"],
            "Name Is"
        ) == (3, 0)
        # An example where doesn't cleanly split, and have
        #   to within one of the parts
        >> find_after_first_occurrence_of_string(
            ["My", " Name", " Is Bob"],
            "Name Is"
        ) == (2j, 3)
        # When it is not found, we return (None, None)
        >> find_after_first_occurrence_of_string(
            ["My", " Name", " Is Bob"],
            "Random"
        ) == (None, None)

    """
    if len(string_parts) == 0:
        return (None, None)

    full_string = "".join(string_parts)  # Concatenate all parts to form the full string
    split_index = full_string.find(
        split_str,
    )  # Find the index of split_str in the full string

    if split_index == -1:
        # If split_str is not found, return (None, None)
        return (None, None)

    # Adjust split_index to point to the character after split_str
    split_index += len(split_str)

    # Now, find which part and the index within that part corresponds to split_index
    current_index = 0
    for i, part in enumerate(string_parts):
        # If the split_index is within the current part
        if current_index + len(part) >= split_index:
            # Calculate the index within the current part
            index_within_part = split_index - current_index
            if index_within_part == len(part):
                # If the index within the part is the length of the part,
                #   it means the split_str is at the end of the part, and
                #   the next part is the one that starts after the split_str
                return (i + 1, 0)
            return (i, index_within_part)
        current_index += len(part)

    msg = "Expected to have returned"
    raise RuntimeError(msg)


def get_only_item(values: Sequence[T] | Iterable[T]) -> T:
    if not isinstance(values, Sequence):
        values = list(values)
    if len(values) != 1:
        msg = f"Expected exactly one item, got {len(values)}"
        raise ValueError(msg)
    return values[0]


def normalize_probs_dict(
    answer_to_prob: dict[str, float],
    available_answers: Sequence[str],
) -> dict[str, float]:
    """
    Makes a new dict that will sum to 1.0. If there are missing
    answers, distributes any available probability mass to missing answers.
    If all the answers are present, then distributes the mass
    in proportion to the current probabilities.
    """
    # Calculate the total probability of the given answers
    total_prob = sum(answer_to_prob.values())

    # Calculate the number of missing answers
    missing_answers = [
        answer for answer in available_answers if answer not in answer_to_prob
    ]
    num_missing_answers = len(missing_answers)
    if 1.0 - 1e-6 <= total_prob <= 1.0 + 1e-6 and num_missing_answers == 0:
        return answer_to_prob

    # Initialize a new dictionary to store the normalized probabilities
    normalized_probs = {}

    # If there are missing answers, distribute the remaining probability mass equally among them
    if num_missing_answers > 0:
        # Calculate the remaining probability mass to distribute
        remaining_prob = 1.0 - total_prob
        # Distribute the remaining probability equally among the missing answers
        for answer in available_answers:
            if answer in answer_to_prob:
                # If the answer is not missing, use its original probability
                normalized_probs[answer] = answer_to_prob[answer]
            else:
                # If the answer is missing, assign it an equal share of the remaining probability
                normalized_probs[answer] = remaining_prob / num_missing_answers
    else:
        # If there are no missing answers, normalize the probabilities so they sum to 1.0
        for answer in available_answers:
            normalized_probs[answer] = answer_to_prob[answer] / total_prob

    # Return in a standard order
    return {answer: normalized_probs[answer] for answer in sorted(available_answers)}


if __name__ == "__main__":
    print("hello world")


import os
from pathlib import Path

import pytest

from synthegrator.lang_specs.lang_spec_python import PythonLangSpec
from synthegrator.memory_fs import ProjectDir, ProjectFile
from synthegrator.sandboxing import (
    PY_DEFAULT_DOCKER_ENV,
    Cmd,
    pytest_on_docker,
    run_on_docker,
    safer_exec, ExecLimits,
)

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


def test_simple_run():
    result = safer_exec("", "3 + 4", user_confirmation=False)
    assert result.output == 7


def test_simple_run_with_code():
    result = safer_exec("foo = lambda x: x*2", "foo(4)", user_confirmation=False)
    assert result.output == 8


def test_with_exit():
    result = safer_exec("exit()", "3 + 4", user_confirmation=False)
    assert result.output is None


def test_with_file_io():
    """Try to write to a temp file. Assert that the file is not written to."""
    import tempfile

    temp_file_name = Path(tempfile.gettempdir()) / "code_for_sandbox_test.txt"
    if temp_file_name.exists():
        temp_file_name.unlink()
    result = safer_exec(
        f"open('{temp_file_name}', 'w').write('hello')",
        "5",
        user_confirmation=False,
    )
    assert result.output is None
    assert result.exception is not None
    assert not temp_file_name.exists()


XML_EXPECTED_EXAMPLE = r"""<?xml version="1.0" encoding="utf-8"?><testsuites name="pytest tests"><testsuite name="pytest" errors="0" failures="1" skipped="0" tests="2" time="0.014" timestamp="2023-07-28T23:17:24.483422" hostname="xxxx"><testcase classname="test_sandboxing" name="test_simple_run" time="0.000" /><testcase classname="test_sandboxing" name="test_simple_run2" time="0.000"><failure message="assert False">def test_simple_run2():
&gt;       assert False
E       assert False

test_sandboxing.py:4: AssertionError</failure></testcase></testsuite></testsuites>"""


def _strip_xml_variance(xml: str) -> str:
    import re

    xml = re.sub(r'timestamp="[^"]+"', "", xml)
    xml = re.sub(r'hostname="[^"]+"', "", xml)
    return re.sub(r'time="[^"]+"', "", xml)


def test_basic_docker_exec():
    r = run_on_docker(
        docker_context=PythonLangSpec().get_default_docker_env(),
        cmds=[
            Cmd("echo hello world"),
        ],
    )
    assert len(r) == 1
    assert r[0].exit_code == 0
    assert r[0].stdout == b"hello world\n"
    assert r[0].completed


def test_basic_docker_exec_with_files():
    r = run_on_docker(
        docker_context=PythonLangSpec().get_default_docker_env(),
        cmds=[
            Cmd("cat foo.txt"),
            Cmd("cat a/bar.txt"),
        ],
        files=ProjectDir(
            {
                "foo.txt": ProjectFile("foo.txt", "hello world"),
                "a": ProjectDir(
                    {
                        "bar.txt": ProjectFile("bar.txt", "bar hello world"),
                        "baz.txt": ProjectFile("baz.txt", "baz hello world"),
                    },
                ),
            },
        ),
    )
    assert len(r) == 2
    assert r[0].exit_code == 0
    assert r[0].stdout == b"hello world"
    assert r[0].completed
    assert r[1].exit_code == 0
    assert r[1].stdout == b"bar hello world"
    assert r[1].completed


def test_pytest_run():
    run_file = "test_sandboxing.py"
    contents = (
        "def test_simple_run():\n"
        "    assert True\n"
        "def test_simple_run2():\n"
        "    assert False\n"
    )
    r = pytest_on_docker(
        files=ProjectDir(
            {
                run_file: ProjectFile(run_file, contents),
            },
        ),
        run_file=run_file,
    )
    assert r.got_results
    assert not r.collection_error
    assert not r.exec_error
    assert r.test_suite_exec_result.exit_code == pytest.ExitCode.TESTS_FAILED
    assert _strip_xml_variance(r.xml_result) == _strip_xml_variance(
        XML_EXPECTED_EXAMPLE,
    )


def test_docker_run_syntax_error():
    run_file = "test_sandboxing.py"
    contents = (
        "foo = 3 +\n"
        "def test_simple_run():\n"
        "    assert True\n"
        "def test_simple_run2():\n"
        "    assert True\n"
    )
    r = pytest_on_docker(
        files=ProjectDir(
            {
                run_file: ProjectFile(run_file, contents),
            },
        ),
        run_file=run_file,
    )
    assert not r.got_results
    assert r.collection_error
    assert not r.exec_error


def test_docker_run_nested_dir():
    r = run_on_docker(
        docker_context=PY_DEFAULT_DOCKER_ENV,
        files=ProjectDir.construct_with_one_file("foo/bar/baz.txt", b"hello world"),
        cmds=[
            "cat foo/bar/baz.txt",
        ],
    )
    assert len(r) == 1
    assert r[0].exit_code == 0
    assert r[0].stdout == b"hello world"


def test_docker_run_timeout():
    """Test that timeouts are properly detected and resources cleaned up."""
    # Create a Python file that contains an infinite loop
    infinite_script = """
import time
while True:
    time.sleep(1)
    print("Still running...")
"""

    files = ProjectDir({
        "infinite_loop.py": ProjectFile("infinite_loop.py", infinite_script)
    })

    # Set a short timeout
    short_limits = ExecLimits(
        timeout_cpu_s=2,
        timeout_realtime_s=2,
        memory_limit_mb=1000,
        max_procs=10  # Allow more processes - important!
    )

    # Run the command
    r = run_on_docker(
        docker_context=PY_DEFAULT_DOCKER_ENV,
        cmds=[Cmd("python infinite_loop.py")],
        files=files,
        limits=short_limits
    )

    # Print diagnostic info
    print(f"Timeout detected: {r[0].timeout}")
    print(f"Completed: {r[0].completed}")
    print(f"Exit code: {r[0].exit_code}")
    print(f"Duration: {r[0].duration}")
    print(f"OOM killed: {r[0].oom_killed}")
    print(f"Stdout: {r[0].stdout}")
    print(f"Stderr: {r[0].stderr}")

    # Assert timeout is correctly detected
    assert r[0].timeout == True
    assert r[0].completed == False


from typing import Generator
from synthegrator.code_problems import CodeProblem
from synthegrator.code_solver import DummyCodeSolverAutoRegressive, LmCodeSolverAutoRegressive
from pathlib import Path
from synthegrator.df_converters import solution_evals_to_df
from synthegrator.execution_threading import solve_and_evaluate_problems
from synthegrator.problem_rendering_insertion_tags import TaggedEditRenderer, TaggedEditResponseParser
from synthegrator.synthdatasets.repocod import yield_repocod_problems
from lmwrapper.openai_wrapper import get_open_ai_lm, OpenAiModelNames
import pandas as pd
from synthegrator.util import IteratorWithLength
import numpy as np

cur_path = Path(__file__).parent


_probs_cache = None

def _get_probs(max_problems = 200) -> IteratorWithLength[CodeProblem]:
    # A lazy cached generator of repocod problems.
    global _probs_cache
    if _probs_cache:
        return IteratorWithLength(iter(_probs_cache), len(_probs_cache))
    
    # Create a generator that caches as it yields
    def cached_generator():
        global _probs_cache
        done_probs = [] 
        for prob in yield_repocod_problems(max_problems=max_problems):
            done_probs.append(prob)
            yield prob
        _probs_cache = done_probs
    
    return IteratorWithLength(cached_generator(), max_problems)



def _make_df(max_problems, model_name, solver_type, cache = True):
    key = f"{model_name}_{solver_type}_{max_problems}"
    save_path = cur_path / f"repocod_evals_{key}.parquet"
    if cache and save_path.exists():
        return pd.read_parquet(save_path), key
    probs = _get_probs(max_problems)
    lm = get_open_ai_lm(
        model_name, 
        #organization=Path("~/decal_oai_org.txt").expanduser().read_text().strip()
    )
    #print(lm.predict("hello world").completion_text)
    #exit()
    if solver_type == "autoregressive":
        solver = LmCodeSolverAutoRegressive(lm)
    elif solver_type == "tagged":
        solver = LmCodeSolverAutoRegressive(
            lm,
            prompt_renderer=TaggedEditRenderer(),
            response_parser=TaggedEditResponseParser(),
        )
    elif solver_type == "dummy":
        solver = DummyCodeSolverAutoRegressive(lm)
    else:
        raise ValueError(f"Unknown solver type: {solver_type}")
    import os
    evals = solve_and_evaluate_problems(
        solver, 
        probs,
        max_threads_eval=max(1, os.cpu_count() // 4),
        max_threads_solve=3,
    )
    #for ev in evals:
    #    print(ev.solution.problem.problem_id)
    #    print(ev.test_results)
    #    if not ev.main_metric.is_success:
    #        for result in ev.test_results:
    #            print(result)
    #    print(ev.main_metric)
    #    exit()
    df = solution_evals_to_df(
        evals, 
        pickle_gzip_whole_solution_eval=False,
    )
    df.to_parquet(cur_path / f"repocod_evals_{key}.parquet", index=False)
    return df, key


def compute_overlap_matrix(all_df):
    """
    Compute overlap matrix showing how many problems pass in common between different keys.
    Returns both count matrix and percentage matrix.
    """
    keys = all_df['key'].unique()
    n_keys = len(keys)
    
    # Get the set of passing problems for each key
    passing_problems = {}
    for key in keys:
        key_df = all_df[all_df['key'] == key]
        passing_probs = set(key_df[key_df['main_metric__is_success']]['solution__problem__problem_id'])
        passing_problems[key] = passing_probs
        print(f"{key}: {len(passing_probs)} passing problems")
    
    # Create overlap matrices
    overlap_counts = np.zeros((n_keys, n_keys), dtype=int)
    overlap_percentages = np.zeros((n_keys, n_keys), dtype=float)
    
    for i, key1 in enumerate(keys):
        for j, key2 in enumerate(keys):
            intersection = passing_problems[key1] & passing_problems[key2]
            union = passing_problems[key1] | passing_problems[key2]
            
            overlap_counts[i, j] = len(intersection)
            
            # Jaccard index: intersection / union
            if len(union) > 0:
                overlap_percentages[i, j] = len(intersection) / len(union)
            else:
                overlap_percentages[i, j] = 0.0
    
    # Create DataFrames for better visualization
    overlap_counts_df = pd.DataFrame(overlap_counts, index=keys, columns=keys)
    overlap_percentages_df = pd.DataFrame(overlap_percentages, index=keys, columns=keys)
    
    return overlap_counts_df, overlap_percentages_df, passing_problems


def analyze_unique_passing_problems(all_df, passing_problems):
    """
    Analyze which problems are uniquely solved by each key.
    """
    keys = all_df['key'].unique()
    
    print("\n=== Unique Passing Problems Analysis ===")
    
    for key in keys:
        key_passing = passing_problems[key]
        # Find problems that only this key solves
        other_keys_passing = set()
        for other_key in keys:
            if other_key != key:
                other_keys_passing.update(passing_problems[other_key])
        
        unique_to_key = key_passing - other_keys_passing
        print(f"\n{key}:")
        print(f"  Total passing: {len(key_passing)}")
        print(f"  Unique to this key: {len(unique_to_key)}")
        
        if len(unique_to_key) > 0:
            print(f"  Unique problem IDs: {sorted(list(unique_to_key))[:5]}{'...' if len(unique_to_key) > 5 else ''}")


def main():
    pass_fracs = {}
    max_problems = 1000
    all_dfs = []
    from datasets import load_dataset
    data = load_dataset('lt-asset/REPOCOD')['test']
    problem_id_to_repo_id = {}
    for sample in data:
        problem_id=f"repocod_{sample['target_module_path']}_{sample['content_class']}_{sample['function_name']}"
        problem_id_to_repo_id[problem_id] = (sample['repo_id'], sample['function_name'], sample['repository'])

    for model_name, solver_type in [
        #(OpenAiModelNames.gpt_4_1, "autoregressive"),
        #(OpenAiModelNames.gpt_4_1_mini, "autoregressive"),
        #(OpenAiModelNames.gpt_4_1, "tagged"),
        #(OpenAiModelNames.gpt_4_1_mini, "tagged"),
        (OpenAiModelNames.gpt_4o_mini, "autoregressive"),
        (OpenAiModelNames.gpt_4_1, "dummy"),
        (OpenAiModelNames.gpt_4_1_mini, "autoregressive"),
        #(OpenAiModelNames.gpt_4_1, "autoregressive"),
    ]:
        df, key = _make_df(max_problems, model_name, solver_type)
        pass_frac = df.main_metric__is_success.mean()
        print(f"Fraction Passing {key}: {pass_frac}")
        pass_fracs[key] = pass_frac
        df["key"] = key
        df["repo_name"] = df.solution__problem__problem_id.apply(lambda x: problem_id_to_repo_id[x][2])
        all_dfs.append(df)
    
    # Compare specific solver types (more robust approach)
    all_df = pd.concat(all_dfs)

    # Configure pandas display options for better readability
    pd.set_option('display.max_columns', None)
    pd.set_option('display.width', None)
    pd.set_option('display.max_colwidth', 50)
    
    # Define aggregation dictionary once for DRY principle
    agg_dict = {
        'main_metric__is_success': ['mean', 'count'],
        'test_results_list__runtime': 'mean',
        'test_results_list__syntax_check_result': 'mean',
        'test_results_list__len': 'mean',
        'test_results_list__collection_error': 'mean',
        'test_results_list__timeout': 'mean'
    }
    
    print("\n=== Summary by Key ===")
    summary_by_key = all_df.groupby("key").agg(agg_dict).round(3)
    
    # Flatten multi-level columns and rename for better readability
    summary_by_key.columns = ['Success_Rate', 'Count', 'Avg_Runtime', 'Syntax_OK', 'Avg_Tests', 'Collection_Error', 'Timeout_Rate']
    print(summary_by_key)
    
    print("\n=== Summary by Key and Repo ===")
    summary_by_key_repo = all_df.groupby(["key", "repo_name"]).agg(agg_dict).round(3)
    
    # Flatten multi-level columns and rename for better readability
    summary_by_key_repo.columns = ['Success_Rate', 'Count', 'Avg_Runtime', 'Syntax_OK', 'Avg_Tests', 'Collection_Error', 'Timeout_Rate']
    print(summary_by_key_repo)

    # Print some of the failing dummy problems
    fail_dummy = all_df[
        (all_df['key'] == 'gpt-4.1_dummy_1000')
        & (all_df['main_metric__is_success'] == False)
    ]
    dummy_fail = "\n".join(list(fail_dummy.solution__problem__problem_id))
    (cur_path / "repocod_gt_fail_synthegrator_flow.txt").write_text(dummy_fail)
    exit()
    
    # Define which solver types we want to compare
    target_keys = {
        'dummy': None,
        'gpt_mini_autoregressive': None
    }
    
    # Find the actual keys that match our target patterns
    for key in all_df['key'].unique():
        if 'dummy' in key:
            target_keys['dummy'] = key
        elif 'gpt-4.1-mini' in key and 'autoregressive' in key:
            target_keys['gpt_mini_autoregressive'] = key
    
    print("target keys:")
    print(target_keys)
    # Check if we found both target solver types
    if target_keys['dummy'] and target_keys['gpt_mini_autoregressive']:
        dummy_key = target_keys['dummy']
        gpt_mini_key = target_keys['gpt_mini_autoregressive']
        
        dummy_df = all_df[all_df['key'] == dummy_key].copy()
        gpt_mini_df = all_df[all_df['key'] == gpt_mini_key].copy()
        
        print(f"\nComparing {dummy_key} vs {gpt_mini_key}")
        print(f"Dummy solver problems: {len(dummy_df)}")
        print(f"GPT mini solver problems: {len(gpt_mini_df)}")
        
        # Quick debug: Show all columns to see what we have
        print(f"\nAll columns in dummy_df: {[col for col in dummy_df.columns if 'solution__' in col][:5]}...")
        print(f"All columns in gpt_mini_df: {[col for col in gpt_mini_df.columns if 'solution__' in col][:5]}...")
        
        # Join on problem_id to compare the same problems
        comparison_df = dummy_df.merge(
            gpt_mini_df, 
            on='solution__problem__problem_id', 
            suffixes=('_dummy', '_gpt_mini')
        )
        
        print(f"Common problems found: {len(comparison_df)}")
        
        # Just show first few comparisons directly
        for i, row in comparison_df.head(20).iterrows():
            dummy_success = row['main_metric__is_success_dummy']
            gpt_success = row['main_metric__is_success_gpt_mini']
            if not gpt_success:
                continue
            problem_id = row['solution__problem__problem_id']
            print(f"\nProblem ID: {problem_id}")
            print("-" * 60)
            
            # Show the prompt from GPT model (dummy doesn't have real prompts)
            #print("PROMPT (from GPT model):")
            #print(row['solution__prompt__text_gpt_mini'][0]['content'])
            #print("\n")
            
            print("DUMMY SOLVER RESPONSE:")
            print(row['solution__solve_steps_dummy'][0]['value'])
            print("\n")
            
            print("GPT-4-1-MINI AUTOREGRESSIVE RESPONSE:")
            print(row['solution__solve_steps_gpt_mini'][0]['value'])
            print("\n")
            
            # Show success comparison
            print(f"Success - Dummy: {dummy_success}, GPT-Mini: {gpt_success}")
            
            print("=" * 80)
    else:
        print(f"\nCould not find both target solver types for comparison:")
        print(f"Available keys: {list(all_df['key'].unique())}")
        print(f"Looking for dummy key: {target_keys['dummy']}")
        print(f"Looking for gpt-mini autoregressive key: {target_keys['gpt_mini_autoregressive']}")
    
    # Continue with rest of analysis if needed
    print(f"\nOverall pass fractions: {pass_fracs}")
    
    # Uncomment this if you want to run the overlap analysis
    # _run_overlap_analysis(all_df)



def _run_overlap_analysis(all_df):
    # Compute and display overlap analysis
    print("\n" + "="*50)
    print("OVERLAP ANALYSIS")
    print("="*50)
    
    overlap_counts, overlap_percentages, passing_problems = compute_overlap_matrix(all_df)
    
    print("\n=== Overlap Counts (Intersection) ===")
    print(overlap_counts)
    
    print("\n=== Overlap Percentages (Jaccard Index) ===")
    print(overlap_percentages.round(3))
    
    analyze_unique_passing_problems(all_df, passing_problems)
    
    # Save the overlap matrices
    overlap_counts.to_csv(cur_path / "overlap_counts.csv")
    overlap_percentages.to_csv(cur_path / "overlap_percentages.csv")
    print(f"\nOverlap matrices saved to {cur_path}")
    



if __name__ == "__main__":
    main()



from synthegrator.memory_fs import ProjectDir, ProjectFile
import pytest
from synthegrator.memory_fs import ProjectDir, ProjectFile


def test_fs():
    mydir = ProjectDir(
        {
            "a": ProjectDir(
                {
                    "b": ProjectFile("b", b"hello"),
                    "c": ProjectFile("c", b"world"),
                },
            ),
            "d": ProjectFile("d", b"!"),
        },
    )
    assert mydir["a"]["b"].content == b"hello"
    assert mydir["a"]["c"].content == b"world"
    assert mydir["d"].content == b"!"
    assert mydir["a"]["b"].name == "b"
    assert mydir["a"]["c"].name == "c"
    assert mydir["d"].name == "d"
    assert mydir["a"].contains_only_leafs()
    assert mydir["a/c"].content == b"world"
    assert not mydir.dirty

    edited = mydir.set_node("d", ProjectFile("d", b"!!", dirty=True))
    assert edited["d"].content == b"!!"
    assert mydir["d"].content == b"!"
    assert edited["a"]["b"].content == b"hello"
    assert not mydir.dirty
    assert edited.dirty
    assert edited["d"].dirty
    assert not edited["a"].dirty

    edited2 = mydir.set_node("a/b", ProjectFile("b", b"hello world"))
    assert edited2["a"]["b"].content == b"hello world"
    assert mydir["a"]["b"].content == b"hello"
    assert edited2["a"]["c"].content == b"world"


def test_fs2():
    mydir = ProjectDir(
        {
            "a": ProjectDir(
                {
                    "b": ProjectFile("b", b"hello"),
                    "c": ProjectFile("c", b"world"),
                },
            ),
            "d": ProjectFile("d", b"!"),
        },
    )
    mydir2 = mydir.set_node("a/b", ProjectFile("b", b"goodbye"))
    assert mydir2["a"]["b"].content == b"goodbye"
    assert mydir["a"]["b"].content == b"hello"


def test_fs3():
    mydir = ProjectDir(
        {
            "a": ProjectDir(
                {
                    "b": ProjectFile("b", b"hello"),
                    "c": ProjectFile("c", b"world"),
                },
            ),
            "d": ProjectFile("d", b"!"),
        },
    )
    mydir2 = ProjectDir.construct_with_files(
        [
            ("a/b", b"hello"),
            ("a/c", b"world"),
            ("d", b"!"),
        ],
    )
    assert mydir2["a"]["b"].content == b"hello"
    assert mydir2["a"]["b"].name == "b"
    assert mydir2["a/b"] == mydir2["a"]["b"]
    assert mydir2["a"]["c"].content == b"world"
    assert mydir == mydir2
    assert mydir.get_hash() == mydir2.get_hash()
    assert mydir != ProjectDir.construct_with_files(
        [
            ("a/b", b"hello"),
        ],
    )


def test_fs4():
    mydir = ProjectDir(
        {
            "a": ProjectDir(
                {
                    "b": ProjectFile("b", b"hello"),
                    "c": ProjectFile("c", b"world"),
                },
            ),
            "d": ProjectFile("d", b"!"),
        },
    )
    assert not mydir.dirty
    mydir = mydir.set_file_contents("a/b", b"goodbye")
    print()
    mydir.pretty_print()
    # fmt: off
    assert mydir.pretty_str() == (
        "a (Dirty)\n"
        "  b (Dirty)\n"
        "  c\n"
        "d\n"
    )
    # fmt: on
    assert mydir.dirty
    assert not mydir["d"].dirty
    assert mydir["a"].dirty
    assert mydir["a"]["b"].dirty
    assert mydir["a"]["b"] == mydir["a/b"]
    assert not mydir["d"].dirty



from dataclasses import dataclass

from synthegrator.memory_fs import ProjectDir
from synthegrator.sandboxing import DockerExecutionContext


@dataclass(frozen=True)
class ProjectWorkingDirectory:
    files: ProjectDir


@dataclass(frozen=True)
class CodeExecutionEnvironment:
    docker_env: DockerExecutionContext = None
    """A docker environment to use when executing the code"""


from typing import Generator, Optional
from pathlib import Path
from datasets import load_dataset

from synthegrator.code_problems import CodeProblem, CodeSolution, DiscoveredTestsuite
from synthegrator.environments import CodeExecutionEnvironment, ProjectWorkingDirectory
from synthegrator.memory_fs import create_project_dir_from_docker_img
from synthegrator.sandboxing import DockerExecutionContext, ExecLimits, run_on_docker
from synthegrator.solution_eval import SolutionEvaluator, evaluate_code_problem_execution, parse_junit_test_cases
from synthegrator.transformation_spec import MarkElement, SolveStep, StsInsert, StsPathSelector, StsSelectorCharSpan, StsSpecStatement, StsTextSelector, StsVerb, TransformationSpec, markup_path
from synthegrator.util import IteratorWithLength

cur_path = Path(__file__).parent


DEFAULT_TIMEOUT_REALTIME_S = 60 * 10
"""10 minute timeout is what they seem to use
https://github.com/lt-asset/REPOCOD/blob/25afbb73672f7a3/evaluate/evaluate.py#L104-L120
"""


def make_docker_env(repo_name: str, timeout_realtime_s: int = DEFAULT_TIMEOUT_REALTIME_S):
    dockerfile = f"""
FROM shanchaol/repocod:{repo_name}
RUN rm -rf /sandbox
RUN ln -s /usr/src/app/{repo_name} /sandbox
# Limit threads to potentially avoid a thread limit error?
ENV OMP_NUM_THREADS=2
ENV OPENBLAS_NUM_THREADS=2
ENV MKL_NUM_THREADS=2
ENV NUMEXPR_NUM_THREADS=2
ENV VECLIB_MAXIMUM_THREADS=2
WORKDIR /sandbox
    """
    return DockerExecutionContext(
        image_name=f"synthegrator__shanchaol/repocod:{repo_name}",
        dockerfile=dockerfile,
        default_limits=ExecLimits(
            memory_limit_mb=16 * 1024,
            timeout_cpu_s=60 * 60 * 5,
            timeout_realtime_s=timeout_realtime_s,
            networking_allowed=False,
        )
    )


def _make_test_run_cmd(repo_name, result_file_name, target_functions_path=None, early_stop=False, n_process = 'auto'):
    # borrowed from repocod author code
    repo_specific_command = ""
    if "xarray" in repo_name:
        repo_specific_command = "/root/miniconda3/bin/conda run -n xarray-tests "
    # Command to run pytest inside the container
    if early_stop:
        if target_functions_path:
            command = f"{repo_specific_command}pytest --continue-on-collection-errors -x --junitxml={result_file_name} -n {n_process} --dist=loadfile -v @{target_functions_path}"
        else:
            command = f"{repo_specific_command}pytest --continue-on-collection-errors -x --junitxml={result_file_name} -n {n_process} --dist=loadfile -v {project_path}"
    else:
        if target_functions_path:
            command = f"{repo_specific_command}pytest --continue-on-collection-errors --junitxml={result_file_name} -n {n_process} --dist=loadfile -v @{target_functions_path}"
        else:
            command = f"{repo_specific_command}pytest --continue-on-collection-errors --junitxml={result_file_name} -n {n_process} --dist=loadfile -v {project_path}"
    return command


def _get_test_case(
    repository: str,
    relavent_test_path: str,
    early_stop: bool = True,
) -> DiscoveredTestsuite:
    # TODO: something weird about scipy
    result_file_name = f"test_results_{repository}.xml"
    return DiscoveredTestsuite(
        cmds=[
            _make_test_run_cmd(
                repo_name=repository,
                result_file_name=result_file_name,
                target_functions_path=relavent_test_path,
                early_stop=early_stop,
                n_process=2,
            ),
            f"cat {result_file_name}",
        ],
    )


_PROJECT_DIR_CACHE = {}
_SOLUTION_EVALUATOR = SolutionEvaluator(
    do_not_execute_if_syntax_fail=False,
)


def _make_problem_id(
    repository: str,
    repo_id: int,
    target_module_path: str,
    relavent_test_path: str,
    prompt: str,
    function_name: str,
    content_class: str,
    full_function: str,
    **kwargs,
):
    problem_id=f"repocod_{target_module_path}_{content_class}_{function_name}"
    return problem_id


def _make_repocod_problem(
    repository: str,
    repo_id: int,
    target_module_path: str,
    relavent_test_path: str,
    prompt: str,
    function_name: str,
    content_class: str,
    full_function: str,
    fail_fast: bool = False,
    timeout_realtime_s: int = DEFAULT_TIMEOUT_REALTIME_S,
):
    test_case = _get_test_case(
        repository=repository,
        relavent_test_path=relavent_test_path,
        early_stop=fail_fast,
    )
    ### Env
    env = make_docker_env(repository, timeout_realtime_s=timeout_realtime_s)
    global _PROJECT_DIR_CACHE
    if env.image_name not in _PROJECT_DIR_CACHE:
        print(f"Building image: {env.image_name}")
        env.build()
        print(f"Build complete, now extracting from: {env.image_name}")
        wd = create_project_dir_from_docker_img(
            env.image_name, 
            f"/usr/src/app/{repository}"
        )
        _PROJECT_DIR_CACHE[env.image_name] = wd
    else:
        wd = _PROJECT_DIR_CACHE[env.image_name]
    assert full_function.startswith(prompt)
    body = full_function[len(prompt):]
    #print("Full function:")
    #print(full_function)
    # Remove the body
    current_text = wd[target_module_path].content_str
    loc = current_text.find(body)
    assert loc != -1
    start = loc
    end = loc + len(body)
    current_text = current_text[:start] + current_text[end:]
    wd = wd.set_file_contents(target_module_path, current_text)
    spec = TransformationSpec([
        StsSpecStatement(
            path_selector=target_module_path,
            text_selector=StsSelectorCharSpan(
                start_char=start,
                end_char=start,
            ),
            verb=StsInsert(
                set_base_space_indent=len(body) - len(body.lstrip(" ")),
            ),
        )
    ])
    instructions = (
        f"The content of the function {function_name} is missing. "
        f"Please complete the body of {function_name}. "
        f"You may only give the body of this function. "
        f"You may not touch any other part of the file. "
        f"(Hint: be careful to match proper indentation of the function)"
    )
    problem = CodeProblem(
        instructions=instructions,
        working_directory=ProjectWorkingDirectory(wd),
        preferred_solution_evaluator=_SOLUTION_EVALUATOR,
        transformation_spec=spec,
        test_cases=[test_case],
        known_solutions=[],
        dataset_name="repocod",
        environment=CodeExecutionEnvironment(
            docker_env=env,
        ),
        problem_id=_make_problem_id(
            **locals(),
        ),
    )
    markup = markup_path(ProjectWorkingDirectory(wd), target_module_path, spec)
    insert_node = next(
        node
        for node in markup.depth_first_iterate()
        if isinstance(node, MarkElement) and isinstance(node.verb, StsInsert)
    )
    known_solution = CodeSolution(
        solve_steps=[
            SolveStep(
                path=target_module_path,
                mark_id=insert_node.mark_id,
                value=body,
            )
        ],
        problem=problem,
    )
    problem.known_solutions.append(known_solution)
    return problem


def yield_repocod_problems(
    max_problems: Optional[int] = None,
    fail_fast: bool = True,
    timeout_realtime_s: int = DEFAULT_TIMEOUT_REALTIME_S,
    filter_out_possible_known_fails: bool = True,
) -> IteratorWithLength[CodeProblem]:
    if filter_out_possible_known_fails:
        known_fails = (cur_path / "repocod_gt_fail_synthegrator_flow.txt").read_text().split("\n")
        known_fails = set(known_fails)
    def _yield_repocod_problems_inner():
        dataset = load_dataset("lt-asset/REPOCOD", streaming=True, split="test")
        for i, sample in enumerate(dataset):
            prob_id = _make_problem_id(
                **sample,
            )
            if filter_out_possible_known_fails and prob_id in known_fails:
                continue
            prob = _make_repocod_problem(
                **sample,
                fail_fast=fail_fast,
                timeout_realtime_s=timeout_realtime_s,
            )
            assert prob_id == prob.problem_id
            yield prob
            if max_problems is not None and i >= max_problems:
                break
    
    return IteratorWithLength(_yield_repocod_problems_inner(), max_problems)


if __name__ == "__main__":
    print("LEN")
    print(len(list(yield_repocod_problems(filter_out_possible_known_fails=True))))
    exit()
    for prob in yield_repocod_problems(timeout_realtime_s=1):
        print(prob)
        solution = prob.known_solutions[0]
        path = "seaborn/_core/plot.py"
        content = solution.apply()[path].content_str
        (cur_path / "dump_t.py").write_text(content)
        result = evaluate_code_problem_execution(prob, solution, do_not_execute_if_syntax_fail=False)
        lang_spec = solution.problem.get_lang_spec_for_path(path)
        syntax_errors = lang_spec.get_syntax_errors(content)
        from pprint import pprint
        pprint(syntax_errors)
        print(result.all_passed())
        print("timeout", result.timeout)
        print("collection_error", result.collection_error)
        print("exec_error", result.exec_error)
        break


import os
import zipfile
from pathlib import Path

import pandas as pd

cur_dir = Path(__file__).parent.absolute()


zip_path = cur_dir / "Defects4J_projects_clean.zip"


def read_file_from_zip(path_in_zip) -> str:
    with zipfile.ZipFile(zip_path, "r") as zip_ref:
        with zip_ref.open(path_in_zip, "r") as file:
            # Read the file content as bytes, then decode to string
            content = file.read().decode("utf-8")
    return content


def loading():
    Path("Defects4J_projects").mkdir(exist_ok=True)
    df = pd.read_csv(
        cur_dir / "Defects4J_oneLiner_metadata.csv",
        header=None,
        names=["Project", "ID", "File_Path", "LOC"],
    )

    DEFECTS4J_COMMAND = "defects4j"
    for _, row in df.iterrows():
        project_name = row["Project"]
        project_id = str(row["ID"])

        buggy_path = f"Defects4J_projects/{project_name}_{project_id}"
        fixed_path = f"{buggy_path}_fixed"

        if not Path(buggy_path).exists():
            os.system(
                f"{DEFECTS4J_COMMAND} checkout -p {project_name} -v {project_id}b -w"
                f" {buggy_path}",
            )

        if not Path(fixed_path).exists():
            os.system(
                f"{DEFECTS4J_COMMAND} checkout -p {project_name} -v {project_id}f -w"
                f" {fixed_path}",
            )

        os.system(f"cp {fixed_path}/.defects4j.config {buggy_path}/.defects4j.config")


if __name__ == "__main__":
    loading()



import itertools
import os
from unittest.mock import MagicMock

import pytest
from lmwrapper.huggingface_wrapper import get_huggingface_lm
from lmwrapper.openai_wrapper import OpenAiModelNames, get_open_ai_lm

from synthegrator.code_solver import (
    DummyCodeSolverAutoRegressive,
    LmCodeSolverAutoRegressive,
)
from synthegrator.df_converters import solution_evals_to_df
from synthegrator.execution_threading import solve_and_evaluate_problems
from synthegrator.synthdatasets.human_eval import yield_human_eval

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"


@pytest.mark.parametrize("max_threads_solve", [1, 2])
@pytest.mark.parametrize("max_threads_eval", [1, 2])
def test_df_converter_dummy(
    max_threads_solve,
    max_threads_eval,
):
    solver = DummyCodeSolverAutoRegressive(MagicMock())
    problems = itertools.islice(yield_human_eval(), 10)
    #problems = list(problems)
    #assert len(problems) == 10
    evals = solve_and_evaluate_problems(
        solver,
        problems,
        max_threads_solve=max_threads_solve,
        max_threads_eval=max_threads_eval,
    )
    df = solution_evals_to_df(evals, solver_key="dummy")
    assert len(df) == 10


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="No api key")
def test_df_converter_gpt():
    lm = get_open_ai_lm(
        OpenAiModelNames.gpt_3_5_turbo,
    )
    solver = LmCodeSolverAutoRegressive(lm)
    problems = itertools.islice(yield_human_eval(), 5)
    evals = solve_and_evaluate_problems(
        solver,
        problems,
        max_threads_solve=1,
        max_threads_eval=1,
    )
    df = solution_evals_to_df(evals, solver_key="gpt35")
    assert len(df) == 5


def test_df_converter_codegen():
    lm = get_huggingface_lm(
        "Salesforce/codegen-350M-multi",
        trust_remote_code=True,
    )
    solver = LmCodeSolverAutoRegressive(lm, include_lm_response=True)
    problems = itertools.islice(yield_human_eval(), 5)
    evals = solve_and_evaluate_problems(
        solver,
        problems,
        max_threads_solve=1,
        max_threads_eval=1,
    )
    df = solution_evals_to_df(evals, solver_key="codegen")
    assert len(df) == 5



import logging

from synthegrator.randomstuff.getk import get_openai_key


def test_oai():
    from langchain.llms import OpenAI

    llm = OpenAI(openai_api_key=get_openai_key())
    logging.info(llm)
    text = "What would be a good company name for a company that makes colorful socks?"
    result = llm.predict(text)
    logging.info(result)
    logging.info(type(result))


def main():
    test_oai()


if __name__ == "__main__":
    main()



import random
import re
from collections.abc import Callable
from dataclasses import dataclass

from synthegrator.code_problems import CodeProblem
from synthegrator.transformation_spec import markup_path


@dataclass
class FewShotLibrary:
    """
    A collections of problems to be used for in-context few-shotting.
    This is a basic implementation that randomly samples
    """

    examples: list[CodeProblem]

    def sample_problems(
        self,
        n: int,
        target_problem: CodeProblem,
        seed: int | None = None,
    ) -> list[CodeProblem]:
        """Sample n problems from the library, excluding the target problem."""
        if seed is None:
            seed = random.randint(0, 2**32 - 1)
        random.seed(seed)
        if n < 0:
            msg = f"n must be non-negative, got {n}"
            raise ValueError(msg)
        if n == 0:
            return []
        n_with_over = min(n + 1, len(self.examples))
        samples: list[CodeProblem] = random.sample(self.examples, n_with_over)
        samples = remove_problem(samples, target_problem)
        return samples[: min(n, len(samples))]


def default_bm25_tokenizer(text: str) -> list[str]:
    # Split based on whitespace
    tokens = re.split(r"\s+", text)

    # Further split based on snake_case and camelCase
    refined_tokens = []
    for token in tokens:
        # Split snake_case
        snake_case_split = re.split(r"_", token)
        for sub_token in snake_case_split:
            # Split camelCase
            camel_case_split = re.findall(
                r"[A-Z]?[a-z]+|[A-Z]+(?=[A-Z][a-z]|$)", sub_token
            )
            if camel_case_split:
                refined_tokens.extend(camel_case_split)
            else:
                refined_tokens.append(sub_token)

    # Split off special symbols from tokens, keeping them as separate tokens
    final_tokens = []
    for token in refined_tokens:
        # Splitting special characters, keeping them as separate tokens
        special_char_split = re.findall(r"[^\w\s]|[\w]+", token)
        final_tokens.extend(special_char_split)

    # Filter out empty strings
    final_tokens = [token for token in final_tokens if token.strip() != ""]

    stop_tokens = {
        " ",
        "\n",
        "\t",
        "[",
        "]",
        "(",
        ")",
        "{",
        "}",
        "<",
        ">",
        "'",
        '"',
        "`",
        ":",
        ";",
        ",",
        ".",
        "?",
        "!",
        "-",
        "_",
        "=",
        "+",
        "*",
        "/",
        "\\",
        "|",
        "&",
        "^",
        "%",
        "$",
        "#",
        "@",
        "~",
    }
    final_tokens = [
        token.lower().strip() for token in final_tokens if token not in stop_tokens
    ]

    return final_tokens


def default_render_func(problem: CodeProblem) -> str:
    from synthegrator.problem_rendering import render_markup_to_text_autoregressive

    instructions = problem.instructions
    texts = [
        render_markup_to_text_autoregressive(
            markup_path(problem.working_directory, path, problem.transformation_spec),
            path,
            problem,
        ).code_prefix
        for path, file in problem.working_directory.files.walk()
    ]
    return "\n".join(texts)


class Bm25FewShotLibrary(FewShotLibrary):
    """
    A few-shot library that uses BM25 to select examples.
    """

    def __init__(
        self,
        examples: list[CodeProblem],
        render_func: Callable[[CodeProblem], str] = default_render_func,
        tokenizer: Callable[[str], list[str]] = default_bm25_tokenizer,
    ):
        self.examples = examples
        try:
            from rank_bm25 import BM25Okapi
        except ImportError:
            raise ImportError(
                "BM25FewShotLibrary requires the rank_bm25 package. "
                "You can install it with `pip install rank-bm25`."
            )
        self.tokenizer = tokenizer
        prompt_texts = [render_func(problem) for problem in examples]
        self.bm25 = BM25Okapi([self.tokenizer(text) for text in prompt_texts])
        self._render_func = render_func

    def sample_problems(
        self,
        n: int,
        target_problem: CodeProblem,
        seed: int | None = None,
    ) -> list[CodeProblem]:
        """Sample n problems from the library, excluding the target problem."""
        if seed is None:
            seed = random.randint(0, 2**32 - 1)
        random.seed(seed)
        if n < 0:
            msg = f"n must be non-negative, got {n}"
            raise ValueError(msg)
        if n == 0:
            return []
        n_with_over = min(n + 1, len(self.examples))
        query = self._render_func(target_problem)
        top_values = self.bm25.get_top_n(
            self.tokenizer(query), self.examples, n_with_over
        )
        top_values = remove_problem(top_values, target_problem)
        top_values = top_values[: min(n, len(top_values))]
        top_values.reverse()  # Most relevant last
        return top_values


def remove_problem(
    problems: list[CodeProblem],
    target_problem: CodeProblem,
) -> list[CodeProblem]:
    # Make sure the target is not included
    return [
        x
        for x in problems
        if (
            x is not target_problem
            # Compare a few key values to exclude if match. Might need
            #    to make more specific later
            and not (
                x.working_directory.files.get_hash()
                == target_problem.working_directory.files.get_hash()
                and x.instructions == target_problem.instructions
                and x.transformation_spec == target_problem.transformation_spec
            )
        )
    ]


@dataclass
class FewShotConfig:
    library: FewShotLibrary
    num_examples: int
    change_examples_between_solves: bool
    """When sampling multiple solutions from a solver,
    should the examples change between each solve?"""



from pathlib import Path

from examples.run_human_eval import run_for_problems_and_serialize
from synthegrator.synthdatasets.mbpp import yield_mbpp

cur_file = Path(__file__).parent.absolute()


def get_mbpp_results(
    output_directory: Path = cur_file / "results",
    read_file_if_exists=False,
    max_limit: int = 500,
):
    print("Gathering all problems...")
    problems = list(
        yield_mbpp(
            sanitized_only=True,
            only_with_passing_reference_solution=True,
            format_as_method_completion_problem=True,
        ),
    )
    return run_for_problems_and_serialize(
        problems,
        "mbpp_func",
        output_directory=output_directory,
        read_file_if_exists=read_file_if_exists,
        max_limit=max_limit,
    )


if __name__ == "__main__":
    get_mbpp_results()



import os
from unittest.mock import MagicMock

import numpy as np
import pytest
from lmwrapper.abstract_predictor import LmPredictor
from lmwrapper.openai_wrapper import (
    OpenAiModelNames,
    get_open_ai_lm,
)
from lmwrapper.structs import LmPrediction, LmPrompt

from synthegrator.code_problem_builders import (
    make_instruction_only_problem,
    make_simple_line_edit_problem,
    make_simple_method_completion_problem,
)
from synthegrator.code_problems import (
    CodeProblem,
    CodeSolution,
    TestCaseMethodCallIsEq,
)
from synthegrator.code_solver import (
    BaseCodeSolver,
    LmCodeSolverAutoRegressive,
    LmUncertaintyHook,
    LmUncertaintyProbAverage,
    LmUncertaintyProbTotal,
)
from synthegrator.few_shotting import FewShotConfig, FewShotLibrary
from synthegrator.lang_specs.lang_spec_python import PythonLangSpec
from synthegrator.lm_few_shotting_tools import FewShotRendererWrapper
from synthegrator.problem_rendering import (
    PromptRendererSingleEditGeneric,
)
from synthegrator.problem_rendering_insertion_tags import (
    TaggedEditRenderer,
    TaggedEditResponseParser,
)
from synthegrator.response_parser import (
    ResponseParserSingleEdit,
    _get_new_content_start,
)
from synthegrator.solution_eval import apply_solution, evaluate_code_problem_execution
from synthegrator.tests.non_dataset_tests.test_problem_rendering import factor_problems
from synthegrator.transformation_spec import (
    StsInsert,
)
from synthegrator.uncertainty_modeling import (
    DelayedUncertaintyEstimate,
    ProbabilityIsCorrect,
)
from synthegrator.util import estimate_best_overlap_span

assert factor_problems, "Unused import detectors struggle with fixtures"

IN_GITHUB_ACTIONS = os.getenv("GITHUB_ACTIONS") == "true"

LM_SOLVERS = [
    LmCodeSolverAutoRegressive,
    # LmCodeSolverAutoRegressive,
]


def run_with_oai(solver: BaseCodeSolver):
    problem = make_simple_method_completion_problem(
        signature="def foo(x) -> int:",
        prompt="Write a function that doubles its input.",
        test_cases=[
            TestCaseMethodCallIsEq(
                test_input=(2,),
                output=4,
                method_name="foo",
                is_hidden_test=False,
                fail_message="",
            ),
            TestCaseMethodCallIsEq(
                test_input=(3,),
                output=6,
                method_name="foo",
                is_hidden_test=False,
                fail_message="",
            ),
        ],
        target_path="solution.py",
    )
    content = problem.working_directory.files["solution.py"].content.decode("utf-8")
    print(content)
    assert (
        content == "def foo(x) -> int:\n"
        '    """\n'
        "    Write a function that doubles its input.\n"
        '    """\n'
    )
    assert problem.working_directory is not None
    assert len(problem.transformation_spec.statements) == 1
    result = solver.solve(problem)
    assert len(result.solve_steps) == 1
    print(result.solve_steps)
    assert result.solve_steps[0].value in {
        "    return x * 2",
        "    return x * 2\n",
        "    return x*2",
        "    return x*2\n",
    }
    evaluations = evaluate_code_problem_execution(
        problem,
        result,
        user_prompted_guard=False,
    )
    assert evaluations.all_passed()


def test_basic_oai_mocked():
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo_instruct)
    lm.predict = MagicMock(return_value=LmPrediction("    return x * 2", None, None))
    run_with_oai(LmCodeSolverAutoRegressive(lm))


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="No API key in GitHub Actions")
def test_basic_oai():
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo_instruct)
    run_with_oai(LmCodeSolverAutoRegressive(lm))


def test_basic_oai_mocked_chat():
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo)
    lm.predict = MagicMock(return_value=LmPrediction("return x * 2", None, None))
    run_with_oai(LmCodeSolverAutoRegressive(lm))


def test_basic_oai_mocked_chat_with_full_method():
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo)
    lm.predict = MagicMock(
        return_value=LmPrediction(
            "def foo(x) -> int:\n"
            '    """\n'
            "    Write a function that doubles its input.\n"
            '    """\n'
            "    return x * 2\n",
            None,
            None,
        ),
    )
    run_with_oai(LmCodeSolverAutoRegressive(lm))


@pytest.mark.parametrize(
    "with_lang_name",
    [True, False],
)
def test_basic_oai_mocked_chat_with_block(with_lang_name: bool):
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo)
    lm.predict = MagicMock(
        return_value=LmPrediction(
            "Sure, here's the solution!\n"
            f"```{'python' if with_lang_name else ''}\n"
            "def foo(x) -> int:\n"
            '    """\n'
            "    Write a function that doubles its input.\n"
            '    """\n'
            "    return x * 2\n"
            "```\n",
            None,
            None,
        ),
    )
    run_with_oai(LmCodeSolverAutoRegressive(lm))


def test_basic_oai_mocked_chat_with_block_no_docstring():
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo)
    lm.predict = MagicMock(
        return_value=LmPrediction(
            "Sure, here's the solution!\n"
            "```python\n"
            "def foo(x) -> int:\n"
            "    return x * 2\n"
            "```\n",
            None,
            None,
        ),
    )
    run_with_oai(LmCodeSolverAutoRegressive(lm))


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="No API key in GitHub Actions")
def test_basic_oai_chat():
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo)
    run_with_oai(LmCodeSolverAutoRegressive(lm))


def run_with_oai_few_shot(lm, factor_problems):
    solver = LmCodeSolverAutoRegressive(
        lm,
        prompt_renderer=PromptRendererSingleEditGeneric(
            few_shot_config_default=FewShotConfig(
                library=FewShotLibrary(factor_problems[:2]),
                num_examples=1,
                change_examples_between_solves=False,
            ),
        ),
    )
    problem = factor_problems[0]
    solution = solver.solve(problem)
    assert solution.solve_steps[0].value in {
        "    return x * 2",
        "    return x * 2\n",
    }


def run_with_oai_few_shot_wrapped(lm, factor_problems):
    renderer = PromptRendererSingleEditGeneric()
    parser = ResponseParserSingleEdit()
    solver = LmCodeSolverAutoRegressive(
        lm,
        prompt_renderer=FewShotRendererWrapper(
            renderer,
            parser,
            few_shot_config=FewShotConfig(
                library=FewShotLibrary(factor_problems[:2]),
                num_examples=1,
                change_examples_between_solves=False,
            ),
        ),
        response_parser=parser,
    )
    problem = factor_problems[0]
    solution = solver.solve(problem)
    assert solution.solve_steps[0].value in {
        "    return x * 2",
        "    return x * 2\n",
    }


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="No API key in GitHub Actions")
def test_chat_few_shot(factor_problems):
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo)
    run_with_oai_few_shot(lm, factor_problems)


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="No API key in GitHub Actions")
def test_chat_few_shot_wrapped(factor_problems):
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo)
    run_with_oai_few_shot_wrapped(lm, factor_problems)


def test_chat_few_shot_mocked(factor_problems):
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo)
    lm.predict = MagicMock(
        return_value=LmPrediction(
            "```\n    return x * 2\n```\n",
            None,
            None,
        ),
    )
    run_with_oai_few_shot(lm, factor_problems)


def test_chat_few_shot_mocked_wrapped(factor_problems):
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo)
    lm.predict = MagicMock(
        return_value=LmPrediction(
            "```\n    return x * 2\n```\n",
            None,
            None,
        ),
    )
    run_with_oai_few_shot_wrapped(lm, factor_problems)


@pytest.mark.parametrize(
    ("first_str", "new_content", "expected"),
    [
        ("hello", "hello world", 5),
        ("abc", "abcd", 3),
        ("abc\n", "\n", 1),
        ("fooabc", "abcd", 3),
        ("", "abcd", 0),
        ("abc", "abcabc", 3),
        ("abcdecd", "cdecdfgh", len("cdecd")),
        ("fooabc", "", 0),
        ("aaacCC", "CCCb", len("CC")),
        # (
        #        "def foo(x) -> int:\n    '''\n    Write a func\n    '''\n",
        #        "def foo(x) -> int:\n    return x * 2\n",
        #        len("def foo() -> int:\n    ")
        # ),
    ],
)
def test_get_new_content_start(first_str, new_content, expected):
    assert _get_new_content_start(first_str, new_content) == expected


@pytest.fixture
def example_edit_problem() -> CodeProblem:
    return make_simple_line_edit_problem(
        code=(
            "def foo(vals: list[int]) -> list[int]:\n"
            '    """\n'
            "    Write a function that doubles every value in a list\n"
            '    """\n'
            "    out = []\n"
            "    for val in vals:\n"
            "        out.append(val * 10)\n"
            "    return out\n"
        ),
        edit_lines_span=(6, 7),
        test_cases=[
            TestCaseMethodCallIsEq(
                test_input=([1, 2, 3],),
                output=[2, 4, 6],
                method_name="foo",
                is_hidden_test=False,
                fail_message="",
            ),
        ],
        current_text_visible=True,
        target_path="solution.py",
        known_solutions=[
            "        out.append(val * 2)\n",
        ],
    )


@pytest.fixture
def example_edit_problem_before_cur_after_invisible() -> CodeProblem:
    return make_simple_line_edit_problem(
        code=(
            "import math\n"
            "\n"
            "def foo(vals: list[int]) -> list[int]:\n"
            '    """\n'
            "    Write a function that doubles every value in a list\n"
            '    """\n'
            "    out = []\n"
            "    for val in vals:\n"
            "        out.append(val * 10)\n"
            "    return out\n"
            "\n"
            "print('hello world')"
        ),
        edit_lines_span=(8, 9),
        invisible_spans=[(0, 2), (10, 12)],
        test_cases=[
            TestCaseMethodCallIsEq(
                test_input=([1, 2, 3],),
                output=[2, 4, 6],
                method_name="foo",
                is_hidden_test=False,
                fail_message="",
            ),
        ],
        current_text_visible=False,
        target_path="solution.py",
        known_solutions=[
            "        out.append(val * 2)\n",
        ],
    )


@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="No API key in GitHub Actions")
def test_edit_solve(example_edit_problem):
    problem = example_edit_problem
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo)
    solver = LmCodeSolverAutoRegressive(lm)
    result = solver.solve(problem)
    assert len(result.solve_steps) == 1
    evaluations = evaluate_code_problem_execution(
        problem,
        result,
        user_prompted_guard=False,
    )
    assert evaluations.all_passed()


@pytest.mark.parametrize("lmsolver", LM_SOLVERS)
@pytest.mark.parametrize(
    "mocked_return",
    [
        "        out.append(val * 2)",
        "        out.append(val * 2)\n",
        "out.append(val * 2)\n",
        "    out.append(val * 2)\n",
        "        out.append(val * 2)\n    return out\n",
        "        out.append(val * 2)\n    return out",
        "        out.append(val * 2) # comment\n    return out # return the output\n",
        "        out.append(val * 2) # comment\n    return out # return the output",
    ],
)
def test_edit_solve_mocked(example_edit_problem, lmsolver, mocked_return):
    # TODO: work even when not forced at 1 line
    problem = example_edit_problem
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo)
    lm.predict = MagicMock(return_value=LmPrediction(mocked_return, None, None))
    solver = lmsolver(lm)
    result = solver.solve(problem)
    assert len(result.solve_steps) == 1
    new_files = apply_solution(result)
    assert (
        new_files.get_only_file().content_str
        == 'def foo(vals: list[int]) -> list[int]:\n    """\n    Write a function'
        ' that doubles every value in a list\n    """\n    out = []\n    for val'
        " in vals:\n        out.append(val *"
        f' 2){" # comment" if " # comment" in mocked_return else ""}\n    return'
        " out\n"
    )
    evaluations = evaluate_code_problem_execution(
        problem,
        result,
        user_prompted_guard=False,
    )
    assert evaluations.all_passed()


@pytest.mark.parametrize("lmsolver", LM_SOLVERS)
@pytest.mark.parametrize(
    "mocked_return",
    [
        "        out.append(val * 2)",
    ],
)
def test_edit_solve_invisible_mocked(
    example_edit_problem_before_cur_after_invisible,
    lmsolver,
    mocked_return,
):
    # TODO: work even when not forced at 1 line
    problem = example_edit_problem_before_cur_after_invisible
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo)
    lm.predict = MagicMock(return_value=LmPrediction(mocked_return, None, None))
    solver = lmsolver(lm)
    result = solver.solve(problem)
    assert len(result.solve_steps) == 1
    new_files = apply_solution(result)
    assert (
        new_files.get_only_file().content_str
        == 'import math\n\ndef foo(vals: list[int]) -> list[int]:\n    """\n   '
        ' Write a function that doubles every value in a list\n    """\n    out'
        " = []\n    for val in vals:\n        out.append(val *"
        f' 2){" # comment" if " # comment" in mocked_return else ""}\n    return'
        " out\n\nprint('hello world')"
    )
    evaluations = evaluate_code_problem_execution(
        problem,
        result,
        user_prompted_guard=False,
    )
    assert evaluations.all_passed()


@pytest.mark.parametrize("lmsolver", LM_SOLVERS)
@pytest.mark.parametrize(
    "model_name",
    [OpenAiModelNames.gpt_3_5_turbo, OpenAiModelNames.gpt_3_5_turbo_instruct],
)
@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="No API key in GitHub Actions")
def test_instructions_only(example_edit_problem, lmsolver, model_name):
    # TODO: work even when not forced at 1 line
    problem = make_instruction_only_problem(
        instructions=(
            "Write a function that doubles its input. "
            "The function should be called `double` and take a single int argument `x`."
        ),
        insert_node=StsInsert(stop_at_block_end=False),
        test_cases=[
            TestCaseMethodCallIsEq(
                is_hidden_test=False,
                fail_message="",
                method_name="double",
                test_input=(x,),
                output=x * 2,
            )
            for x in range(3)
        ],
        known_solutions=["def double(x: int) -> int:\n    return x * 2"],
    )
    evaluations = evaluate_code_problem_execution(
        problem,
        problem.known_solutions[0],
    )
    assert evaluations.all_passed()
    solver = lmsolver(get_open_ai_lm(model_name))
    solve = solver.solve(problem)
    print("Solution:")
    print(apply_solution(solve).get_only_file().content_str)
    eval = evaluate_code_problem_execution(
        problem,
        solve,
    )
    assert eval.all_passed()


@pytest.mark.skip(reason="Uncertainty hooks needs to be refactored")
@pytest.mark.skipif(IN_GITHUB_ACTIONS, reason="No API key in GitHub Actions")
def test_uncertainty_hook(factor_problems):
    lm = get_open_ai_lm(OpenAiModelNames.gpt_3_5_turbo_instruct)
    solver = LmCodeSolverAutoRegressive(lm)
    solver.add_uncertainty_hook(LmUncertaintyProbAverage())
    double_prob = factor_problems[0]
    solution = solver.solve(double_prob)
    assert len(solution.uncertainty_estimates) == 1
    file = apply_solution(solution).get_only_file().content_str
    estimate = solution.uncertainty_estimates[0].get_estimate()
    assert isinstance(estimate, ProbabilityIsCorrect)
    pred = lm.predict(LmPrompt(file, echo=True, max_tokens=0))
    print(pred.get_full_tokens())
    prompt_tokens = [
        "def",
        " double",
        "(",
        "x",
        ")",
        " ->",
        " int",
        ":",
        "\n",
        "   ",
        ' """',
        "\n",
        "   ",
        " Write",
        " a",
        " function",
        " that",
        " doubles",
        " its",
        " input",
        ".",
        "\n",
        "   ",
        ' """',
        "\n",
    ]
    continue_tokens = pred.get_full_tokens()[len(prompt_tokens) :]
    print(continue_tokens)
    log_probs = pred.full_logprobs[len(prompt_tokens) :]
    print(log_probs)
    avg = np.average(np.exp(log_probs))
    print(avg)
    assert avg == pytest.approx(estimate.value, abs=0.005)


@pytest.mark.parametrize("lmsolver", LM_SOLVERS)
@pytest.mark.skip(reason="Uncertainty hooks needs to be refactored")
def test_block_end(lmsolver):
    lm = get_open_ai_lm()
    pred = MagicMock()
    pred.completion_text = "    return 2\nassert False"
    pred.completion_tokens = ["   ", " return", " 2", "\n", "assert", " False"]
    completion_probs = [-0.1, -0.2, -0.3, -0.4, -0.5, -0.6]
    pred.completion_logprobs = completion_probs
    lm.predict = MagicMock(return_value=pred)
    solver = lmsolver(lm)
    num_calls = 0

    class ToyHook(LmUncertaintyHook):
        def prepare(
            self,
            lm_prediction: LmPrediction,
            model: LmPredictor,
            solution: CodeSolution,
            modified_full_text: str,
        ) -> DelayedUncertaintyEstimate:
            nonlocal num_calls
            num_calls += 1
            assert len(solution.solve_steps) == 1
            assert estimate_best_overlap_span(
                lm_prediction.completion_text,
                larger_tokens=lm_prediction.completion_tokens,
                mod_string=solution.solve_steps[0].value,
                tokenizer_func=model.tokenize,
            )
            if lmsolver == LmCodeSolverAutoRegressive:
                assert modified_full_text == (  # fmt: off
                    #'Complete the following method\n'
                    #'\n'
                    'def foo():\n    """\n    Always return 2\n    """\n    return 2'
                )  # fmt: on

    solver.add_uncertainty_hook(ToyHook())
    problem = make_simple_method_completion_problem(
        prompt="Always return 2",
        signature="def foo():",
        lang_spec=PythonLangSpec(),
        test_cases=[],
    )
    assert not problem.instructions_are_essential
    solution = solver.solve(problem)
    assert num_calls == 1
    content = apply_solution(solution).get_only_file().content_str
    assert content == (  # fmt: off
        'def foo():\n    """\n    Always return 2\n    """\n    return 2'
    )  # fmt: on


@pytest.mark.parametrize("lmsolver", LM_SOLVERS)
@pytest.mark.skip(reason="Uncertainty hooks needs to be refactored")
def test_block_end_avg(lmsolver):
    lm = get_open_ai_lm()
    pred = MagicMock()
    pred.completion_text = "    return 2\nassert False"
    pred.completion_tokens = ["   ", " return", " 2", "\n", "assert", " False"]

    def mock_tokenize(input_str):
        if input_str == pred.completion_text:
            return pred.completion_tokens
        if input_str == "    return 2\n":
            return ["   ", " return", " 2", "\n"]
        if input_str == "    return 2":
            return ["   ", " return", " 2"]
        msg = "Unexpected tokenizing input"
        raise AssertionError(msg)

    lm.tokenize = MagicMock(side_effect=mock_tokenize)
    completion_probs = [-0.1, -0.2, -0.3, -0.4, -0.5, -0.6]
    pred.completion_logprobs = completion_probs
    lm.predict = MagicMock(return_value=pred)
    solver = lmsolver(lm)
    solver.add_uncertainty_hook(LmUncertaintyProbTotal())
    problem = make_simple_method_completion_problem(
        prompt="Always return 2",
        signature="def foo():",
        lang_spec=PythonLangSpec(),
        test_cases=[],
    )
    solution = solver.solve(problem)
    val = solution.uncertainty_estimates[0].get_estimate().value
    expected = np.exp(np.sum(np.array(completion_probs)[:3]))
    assert val == pytest.approx(expected)


@pytest.mark.parametrize("lmsolver", LM_SOLVERS)
@pytest.mark.skip(reason="Uncertainty hooks needs to be refactored")
def test_block_end2(lmsolver):
    lm = get_open_ai_lm()
    pred = MagicMock()
    pred.completion_text = "    return 2\nassert False"
    pred.completion_tokens = ["   ", " return", " 2", "\n", "assert", " False"]

    def mock_tokenize(input_str):
        if input_str == pred.completion_text:
            return pred.completion_tokens
        if input_str == "    return 2\n":
            return ["   ", " return", " 2", "\n"]
        if input_str == "    return 2":
            return ["   ", " return", " 2"]
        msg = "Unexpected tokenizing input"
        raise AssertionError(msg)

    lm.tokenize = MagicMock(side_effect=mock_tokenize)
    completion_probs = [-0.1, -0.2, -0.3, -0.4, -0.5, -0.6]
    pred.completion_logprobs = completion_probs
    lm.predict = MagicMock(return_value=pred)
    solver = lmsolver(lm)
    solver.add_uncertainty_hook(LmUncertaintyProbAverage())
    problem = make_simple_method_completion_problem(
        prompt="Always return 2",
        signature="def foo():",
        lang_spec=PythonLangSpec(),
        test_cases=[],
    )
    solution = solver.solve(problem)
    val = solution.uncertainty_estimates[0].get_estimate().value
    expected = np.average(np.exp(np.array(completion_probs)[:3]))
    assert val == pytest.approx(expected)


@pytest.mark.parametrize("lmsolver", LM_SOLVERS)
def test_lead_new_line(lmsolver):
    lm = get_open_ai_lm()
    pred = MagicMock()
    pred.completion_text = "\n  return 2"
    lm.predict = MagicMock(return_value=pred)
    solver = lmsolver(lm)
    problem = make_simple_method_completion_problem(
        prompt="Always return 2",
        signature="def foo():",
        lang_spec=PythonLangSpec(),
        test_cases=[],
    )
    solution = solver.solve(problem)
    assert solution.solve_steps[0].value == "    return 2"
    apply_solution(solution)


@pytest.mark.parametrize("lmsolver", LM_SOLVERS)
def test_lead_new_line2(lmsolver):
    lm = get_open_ai_lm()
    pred = MagicMock()
    pred.completion_text = "\nreturn 2"
    lm.predict = MagicMock(return_value=pred)
    solver = lmsolver(lm)
    problem = make_simple_method_completion_problem(
        prompt="Always return 2",
        signature="def foo():",
        lang_spec=PythonLangSpec(),
        test_cases=[],
    )
    solution = solver.solve(problem)
    assert solution.solve_steps[0].value == "    return 2"
    apply_solution(solution)


# fmt: off
@pytest.mark.parametrize(
    "lm_name",
    [
        OpenAiModelNames.gpt_3_5_turbo_instruct,
        OpenAiModelNames.gpt_4,
    ],
)
@pytest.mark.parametrize(
    "completion_text",
    [
        '<edit_solve id="0_qkIO0">    return x * 2</edit_solve>',
        '<edit_solve id="0_qkIO0">return x * 2</edit_solve>',
        (
            "The answer is"
            '\n<edit_solve id="0_qkIO0">    return x * 2</edit_solve>\n'
            "Let me know what else you need"
        ),
        "<edit_solve>return x * 2</edit_solve>",  # Default value
        "return x * 2</edit_solve>",
        "    return x * 2",
        (
            "```python\n"
            "<edit_solve>     return x * 2</edit_solve>\n"
            "```"
        ),
    ],
)
def test_edit_solver(factor_problems, lm_name, completion_text):
    # fmt: on
    lm = get_open_ai_lm(lm_name)
    problem = factor_problems[0]
    pred = MagicMock()
    pred.completion_text = completion_text
    lm.predict = MagicMock(return_value=pred)
    renderer = TaggedEditRenderer()
    render = renderer(problem, MagicMock())
    parser = TaggedEditResponseParser()
    assert parser.parse(render, pred, problem)[0].value == "    return x * 2"
    solver = LmCodeSolverAutoRegressive(
        model=lm,
        prompt_renderer=renderer,
        response_parser=parser,
    )
    solution = solver.solve(problem)
    assert len(solution.solve_steps) == 1
    assert solution.solve_steps[0].value == "    return x * 2"
    apply_solution(solution)



@pytest.mark.parametrize(
    "lm_name",
    [
        OpenAiModelNames.gpt_3_5_turbo_instruct,
        OpenAiModelNames.gpt_4,
    ],
)
@pytest.mark.parametrize(
    "completion_text",
    [
        (
            "```java\n"
            "<fix>    if (options.dependencyOptions.needsManagement() && options.shouldRunClosurePass()) {</fix>\n"
            "```"
        )
    ],
)
def test_edit_solver(factor_problems, lm_name, completion_text):
    lm = get_open_ai_lm(lm_name)
    problem = factor_problems[0]
    pred = MagicMock()
    pred.completion_text = completion_text
    lm.predict = MagicMock(return_value=pred)
    renderer = TaggedEditRenderer(
        tag_name_edit="buggy",
        tag_name_solve="fix",
        include_first_tag_at_end=None,
        custom_closing_lines=(
            "Question: There is a bug in the above code snippet tagged by <buggy> and </buggy>. "
            "Please generate the correct version inside of tags <fix> and </fix>.\n"
            "Answer:\n"
        ),
    )
    render = renderer(problem, MagicMock())
    parser = TaggedEditResponseParser()
    v = "    if (options.dependencyOptions.needsManagement() && options.shouldRunClosurePass()) {"
    result = parser.parse(render, pred, problem)[0].value
    print("Expected: ", repr(v))
    print("Result: ", repr(result))
    assert result == v
    solver = LmCodeSolverAutoRegressive(
        model=lm,
        prompt_renderer=renderer,
        response_parser=parser,
    )
    solution = solver.solve(problem)
    assert len(solution.solve_steps) == 1
    assert solution.solve_steps[0].value == v
    apply_solution(solution)
